from __future__ import annotations

import pandas as pd
import pytest
from sklearn import naive_bayes as sk_naive_bayes

from river import compose, feature_extraction, naive_bayes


def river_models():
    """List of Naive Bayes models to test."""
    yield from [
        naive_bayes.MultinomialNB,
        naive_bayes.BernoulliNB,
        naive_bayes.ComplementNB,
    ]


def sklearn_models():
    """Mapping between Naives Bayes river models and sklearn models."""
    yield from [
        (naive_bayes.MultinomialNB, sk_naive_bayes.MultinomialNB),
        (naive_bayes.BernoulliNB, sk_naive_bayes.BernoulliNB),
        (naive_bayes.ComplementNB, sk_naive_bayes.ComplementNB),
    ]


def yield_dataset():
    """Incremental dataset."""
    yield from [
        ("Chinese Beijing Chinese", "yes"),
        ("Chinese Chinese Shanghai", "yes"),
        ("Chinese Macao", "yes"),
        ("Tokyo Japan Chinese", "no"),
    ]


def yield_batch_dataset():
    """Batch dataset."""
    for x, y in yield_dataset():
        yield pd.Series([x]), pd.Series([y])


def yield_unseen_data():
    yield from [
        "Chinese Beijing Chinese",
        "Chinese Chinese Shanghai",
        "Chinese Macao",
        "Tokyo Japan Chinese",
        "new unseen data",
        "Taiwanese Taipei",
        "Chinese Shanghai" "Shanghai",
        "Chinese",
        "Tokyo Macao",
        "Tokyo Tokyo",
        "Macao Macao new",
        "new",
    ]


def yield_batch_unseen_data():
    yield from [pd.Series(x) for x in yield_unseen_data()]


@pytest.mark.parametrize(
    "model",
    [
        pytest.param(
            compose.Pipeline(
                ("tokenize", feature_extraction.BagOfWords(lowercase=False)),
                ("model", model(alpha=alpha)),
            ),
            id=f"{model.__name__} - {alpha}",
        )
        for model in river_models()
        for alpha in [alpha for alpha in range(1, 4)]
    ],
)
def test_learn_one_methods(model):
    """Assert that the methods of the Naives Bayes class behave correctly."""
    assert model.predict_proba_one("not fitted yet") == {}
    assert model.predict_one("not fitted yet") is None

    for x, y in yield_dataset():
        model.learn_one(x, y)

    # Check class methods.
    if isinstance(model["model"], naive_bayes.ComplementNB) or isinstance(
        model["model"], naive_bayes.MultinomialNB
    ):
        assert model["model"]._more_tags() == {"positive input"}
    assert model["model"]._multiclass


@pytest.mark.parametrize(
    "model, batch_model",
    [
        pytest.param(
            compose.Pipeline(
                ("tokenize", feature_extraction.BagOfWords(lowercase=False)),
                ("model", model(alpha=alpha)),
            ),
            compose.Pipeline(
                ("tokenize", feature_extraction.BagOfWords(lowercase=False)),
                ("model", model(alpha=alpha)),
            ),
            id=f"{model.__name__} - {alpha}",
        )
        for model in river_models()
        for alpha in [alpha for alpha in range(1, 4)]
    ],
)
def test_learn_many_vs_learn_one(model, batch_model):
    """Assert that the Naive Bayes river models provide the same results when learning in
    incremental and mini-batch modes. The models tested are MultinomialNB, BernoulliNB and
    ComplementNB with differents alpha parameters..
    """
    for x, y in yield_dataset():
        model.learn_one(x, y)

    for x, y in yield_batch_dataset():
        batch_model.learn_many(x, y)

    assert model["model"].p_class("yes") == batch_model["model"].p_class("yes")
    assert model["model"].p_class("no") == batch_model["model"].p_class("no")

    # Assert batch and incremental models give same results
    for x, x_batch in zip(yield_unseen_data(), yield_batch_unseen_data()):
        assert model.predict_proba_one(x)["yes"] == pytest.approx(
            batch_model.predict_proba_many(x_batch)["yes"][0]
        )

        assert model.predict_proba_one(x)["no"] == pytest.approx(
            batch_model.predict_proba_many(x_batch)["no"][0]
        )

    # Assert class probabilities are the same when trainig Naive Bayes in pure online and in batch.
    assert model["model"].p_class("yes") == batch_model["model"].p_class("yes")
    assert model["model"].p_class("no") == batch_model["model"].p_class("no")

    # Assert conditionnal probabilities are the same when training Naive Bayes in pure online and
    # in batch.
    if isinstance(model["model"], naive_bayes.BernoulliNB) or isinstance(
        model["model"], naive_bayes.MultinomialNB
    ):
        inc_cp = model["model"].p_feature_given_class
        batch_cp = batch_model["model"].p_feature_given_class

        assert inc_cp("Chinese", "yes") == batch_cp("Chinese", "yes")
        assert inc_cp("Tokyo", "yes") == batch_cp("Tokyo", "yes")
        assert inc_cp("Japan", "yes") == batch_cp("Japan", "yes")
        assert inc_cp("Chinese", "no") == batch_cp("Chinese", "no")
        assert inc_cp("Tokyo", "no") == batch_cp("Tokyo", "no")
        assert inc_cp("Japan", "no") == batch_cp("Japan", "no")
        assert inc_cp("unseen", "yes") == batch_cp("unseen", "yes")

    assert model["model"].class_counts == batch_model["model"].class_counts
    assert model["model"].feature_counts == batch_model["model"].feature_counts

    if isinstance(model["model"], naive_bayes.ComplementNB) or isinstance(
        model["model"], naive_bayes.MultinomialNB
    ):
        assert model["model"].class_totals == batch_model["model"].class_totals

    if isinstance(model["model"], sk_naive_bayes.ComplementNB):
        assert model["model"].feature_totals == batch_model["model"].feature_totals


@pytest.mark.parametrize(
    "batch_model",
    [
        pytest.param(
            compose.Pipeline(
                ("tokenize", feature_extraction.BagOfWords(lowercase=False)),
                ("model", model(alpha=alpha)),
            ),
            id=f"{model.__name__} - {alpha}",
        )
        for model in river_models()
        for alpha in [alpha for alpha in range(1, 4)]
    ],
)
def test_learn_many_not_fit(batch_model):
    """Ensure that Naives Bayes models return an empty DataFrame when not yet fitted. Also check
    that the predict_proba_many method keeps the index.
    """
    assert batch_model.predict_proba_many(
        pd.Series(["new", "unseen"], index=["river", "rocks"])
    ).equals(pd.DataFrame(index=["river", "rocks"]))

    assert batch_model.predict_many(pd.Series(["new", "unseen"], index=["river", "rocks"])).equals(
        pd.DataFrame(index=["river", "rocks"])
    )


@pytest.mark.parametrize(
    "model, sk_model, bag",
    [
        pytest.param(
            compose.Pipeline(
                ("tokenize", feature_extraction.BagOfWords(lowercase=False)),
                ("model", model(alpha=alpha)),
            ),
            sk_model(alpha=alpha),
            feature_extraction.BagOfWords(lowercase=False),
            id=f"{model.__name__} - {alpha}",
        )
        for model, sk_model in sklearn_models()
        for alpha in [alpha for alpha in range(1, 4)]
    ],
)
def test_river_vs_sklearn(model, sk_model, bag):
    """Assert that river Naive Bayes models and sklearn Naive Bayes models provide the same results
    when the input data are the same. Also check that the behaviour of Naives Bayes models are the
    same with dense and sparse dataframe. Models tested are MultinomialNB, BernoulliNB and
    ComplementNB with differents alpha parameters.
    """
    for x, y in yield_batch_dataset():
        model.learn_many(x, y)

    X = pd.concat([x for x, _ in yield_batch_dataset()])
    y = pd.concat([y for _, y in yield_batch_dataset()])

    sk_model = sk_model.fit(X=bag.transform_many(X), y=y)

    # Assert river produce same results as sklearn using sparse dataframe:
    for sk_preds, river_preds in zip(
        sk_model.predict_proba(bag.transform_many(X)),
        model.predict_proba_many(X).values,
    ):
        for sk_pred, river_pred in zip(sk_preds, river_preds):
            assert river_pred == pytest.approx(1 - sk_pred) or river_pred == pytest.approx(sk_pred)

    # Assert river produce same results as sklearn using dense dataframe:
    for sk_preds, river_preds in zip(
        sk_model.predict_proba(bag.transform_many(X).sparse.to_dense()),
        model["model"].predict_proba_many(bag.transform_many(X).sparse.to_dense()).values,
    ):
        for sk_pred, river_pred in zip(sk_preds, river_preds):
            assert river_pred == pytest.approx(1 - sk_pred) or river_pred == pytest.approx(sk_pred)



from __future__ import annotations

import random

from river import stats


def test_issue_1178():
    """

    https://github.com/online-ml/river/issues/1178

    >>> from river import stats

    >>> q = stats.Quantile(0.01)
    >>> for x in [5, 0, 0, 0, 0, 0, 0, 0]:
    ...     q.update(x)
    ...     print(q)
    Quantile: 5.
    Quantile: 0.
    Quantile: 0.
    Quantile: 0.
    Quantile: 0.
    Quantile: 0.
    Quantile: 0.
    Quantile: 0.

    >>> q = stats.Quantile(0.99)
    >>> for x in [5, 0, 0, 0, 0, 0, 0, 0]:
    ...     q.update(x)
    ...     print(q)
    Quantile: 5.
    Quantile: 5.
    Quantile: 5.
    Quantile: 5.
    Quantile: 5.
    Quantile: 0.
    Quantile: 0.277778
    Quantile: 0.827546

    """


def test_ge():
    low = stats.Quantile(0.01)
    high = stats.Quantile(0.99)

    for _ in range(100):
        x = random.random()
        low.update(x)
        high.update(x)
        assert high.get() >= low.get()



from __future__ import annotations

from river.tree.base import Branch, Leaf


class BinaryBranch(Branch):
    def __init__(self, left, right, threshold=None, **kwargs):
        super().__init__(left, right)
        self.threshold = threshold
        self.__dict__.update(kwargs)

    def next(self, x):
        if x < self.threshold:
            return self.children[0]
        return self.children[1]

    def most_common_path(self):
        raise NotImplementedError

    @property
    def repr_split(self):
        return f"x < {self.threshold}"


def test_size():
    tree = BinaryBranch(
        BinaryBranch(
            BinaryBranch(BinaryBranch(Leaf(), Leaf()), Leaf()),
            BinaryBranch(Leaf(), Leaf()),
        ),
        BinaryBranch(Leaf(), Leaf()),
    )

    assert tree.n_nodes == tree.n_branches + tree.n_leaves == 6 + 7
    assert (
        tree.children[0].n_nodes == tree.children[0].n_branches + tree.children[0].n_leaves == 4 + 5
    )
    assert (
        tree.children[1].n_nodes == tree.children[1].n_branches + tree.children[1].n_leaves == 1 + 2
    )
    assert (
        tree.children[1].children[0].n_nodes
        == tree.children[1].children[0].n_branches + tree.children[1].children[0].n_leaves
        == 0 + 1
    )


def test_height():
    tree = BinaryBranch(
        BinaryBranch(
            BinaryBranch(
                BinaryBranch(Leaf(), Leaf()),
                Leaf(),
            ),
            BinaryBranch(Leaf(), Leaf()),
        ),
        BinaryBranch(Leaf(), Leaf()),
    )

    assert tree.height == 5
    assert tree.children[0].height == 4
    assert tree.children[1].height == 2
    assert tree.children[1].children[0].height == 1


def test_iter_dfs():
    tree = BinaryBranch(BinaryBranch(Leaf(no=3), Leaf(no=4), no=2), Leaf(no=5), no=1)

    for i, node in enumerate(tree.iter_dfs(), start=1):
        assert i == node.no


def test_iter_bfs():
    tree = BinaryBranch(BinaryBranch(Leaf(no=4), Leaf(no=5), no=2), Leaf(no=3), no=1)

    for i, node in enumerate(tree.iter_bfs(), start=1):
        assert i == node.no


def test_iter_leaves():
    tree = BinaryBranch(BinaryBranch(Leaf(no=1), Leaf(no=2)), Leaf(no=3))

    for i, leaf in enumerate(tree.iter_leaves(), start=1):
        assert i == leaf.no


def test_iter_branches():
    tree = BinaryBranch(
        BinaryBranch(BinaryBranch(Leaf(), Leaf(), no=3), Leaf(), no=2),
        BinaryBranch(Leaf(), Leaf(), no=4),
        no=1,
    )

    for i, branch in enumerate(tree.iter_branches(), start=1):
        assert i == branch.no


def test_iter_edges():
    tree = BinaryBranch(BinaryBranch(Leaf(no=3), Leaf(no=4), no=2), Leaf(no=5), no=1)

    order = [(1, 2), (2, 3), (2, 4), (1, 5)]

    for i, (parent, child) in enumerate(tree.iter_edges()):
        assert order[i] == (parent.no, child.no)



from __future__ import annotations

import numpy as np

from river import utils


def test_dotvecmat_zero_vector_times_matrix_of_ones():
    A_numpy = np.array([[1, 1], [1, 1]])
    A_river = {(0, 0): 1, (0, 1): 1, (1, 0): 1, (1, 1): 1}
    x_numpy = np.array([0, 0])
    x_river = {0: 0, 1: 0}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]


def test_dotvecmat_vector_of_ones_times_zero_matrix():
    A_numpy = np.array([[0, 0], [0, 0]])
    A_river = {(0, 0): 0, (0, 1): 0, (1, 0): 0, (1, 1): 0}
    x_numpy = np.array([1, 1])
    x_river = {0: 1, 1: 1}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]


def test_dotvecmat_filter_first_matrix_row_with_vector():
    A_numpy = np.array([[1, 1], [1, 1]])
    A_river = {(0, 0): 1, (0, 1): 1, (1, 0): 1, (1, 1): 1}
    x_numpy = np.array([1, 0])
    x_river = {0: 1, 1: 0}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]


def test_dotvecmat_filter_second_matrix_row_with_vector():
    A_numpy = np.array([[1, 1], [1, 1]])
    A_river = {(0, 0): 1, (0, 1): 1, (1, 0): 1, (1, 1): 1}
    x_numpy = np.array([0, 1])
    x_river = {0: 0, 1: 1}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]


def test_dotvecmat_vector_times_identity_matrix():
    A_numpy = np.array([[1, 0], [0, 1]])
    A_river = {(0, 0): 1, (0, 1): 0, (1, 0): 0, (1, 1): 1}
    x_numpy = np.array([1, 1])
    x_river = {0: 1, 1: 1}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]


def test_dotvecmat_vector_times_anti_diagonal_identity_matrix():
    A_numpy = np.array([[0, 1], [1, 0]])
    A_river = {(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 0}
    x_numpy = np.array([1, 1])
    x_river = {0: 1, 1: 1}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]


def test_dotvecmat_three_dimensional_vector_times_non_quadratic_matrix():
    A_numpy = np.array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10], [11, 12, 13, 14, 15]])
    A_river = {
        (0, 0): 1,
        (0, 1): 2,
        (0, 2): 3,
        (0, 3): 4,
        (0, 5): 5,
        (1, 0): 6,
        (1, 1): 7,
        (1, 2): 8,
        (1, 3): 9,
        (1, 5): 10,
        (2, 0): 11,
        (2, 1): 12,
        (2, 2): 13,
        (2, 3): 14,
        (2, 5): 15,
    }
    x_numpy = np.array([1, 2, 3])
    x_river = {0: 1, 1: 2, 2: 3}

    numpy_dotvecmat = x_numpy.dot(A_numpy)
    river_dotvecmat = utils.math.dotvecmat(x_river, A_river)

    assert numpy_dotvecmat[0] == river_dotvecmat[0]
    assert numpy_dotvecmat[1] == river_dotvecmat[1]



from __future__ import annotations

from river import datasets, evaluate, linear_model, metrics, optim, preprocessing


def test_issue_1443():
    dataset = datasets.Phishing()

    model = preprocessing.StandardScaler() | linear_model.LogisticRegression(
        optimizer=optim.SGD(0.1)
    )

    metric = metrics.ConfusionMatrix()

    for _ in evaluate.iter_progressive_val_score(dataset, model, metric):
        pass


def test_confusion_and_other_metrics():
    """

    >>> dataset = datasets.Phishing()

    >>> model = preprocessing.StandardScaler() | linear_model.LogisticRegression(
    ...     optimizer=optim.SGD(0.1)
    ... )

    >>> metric = metrics.ConfusionMatrix() + metrics.F1() + metrics.Accuracy()

    >>> evaluate.progressive_val_score(dataset, model, metric)
            False   True
    False     613     89
     True      49    499
    F1: 87.85%
    Accuracy: 88.96%

    """



from __future__ import annotations

import pytest

from river import datasets, tree
from river.datasets import synth


def get_regression_data():
    return iter(synth.Friedman(seed=42).take(200))


@pytest.mark.parametrize(
    "dataset, splitter",
    [
        (datasets.Phishing(), tree.splitter.ExhaustiveSplitter()),
        (datasets.Phishing(), tree.splitter.HistogramSplitter()),
        (datasets.Phishing(), tree.splitter.GaussianSplitter()),
    ],
)
def test_class_splitter(dataset, splitter):
    model = tree.HoeffdingTreeClassifier(
        splitter=splitter, grace_period=10, leaf_prediction="mc", delta=0.1
    )

    for x, y in dataset:
        model.learn_one(x, y)

    assert model.height > 0


@pytest.mark.parametrize(
    "dataset, splitter",
    [
        (get_regression_data(), tree.splitter.EBSTSplitter()),
        (get_regression_data(), tree.splitter.TEBSTSplitter()),
        (get_regression_data(), tree.splitter.QOSplitter()),
        (get_regression_data(), tree.splitter.QOSplitter(allow_multiway_splits=True)),
    ],
)
def test_reg_splitter(dataset, splitter):
    model = tree.HoeffdingTreeRegressor(
        splitter=splitter, grace_period=20, delta=0.1, leaf_prediction="mean"
    )

    for x, y in dataset:
        model.learn_one(x, y)

    assert model.height > 0


def test_nominal_reg_splitter():
    dataset = synth.Mv(seed=42).take(200)
    model = tree.HoeffdingTreeRegressor(grace_period=10, leaf_prediction="mean")

    for x, y in dataset:
        model.learn_one(x, y)

    assert model.height > 0

    # Evaluates nominal binary splits
    dataset = synth.Mv(seed=42).take(200)
    model = tree.HoeffdingTreeRegressor(grace_period=10, leaf_prediction="mean", binary_split=True)

    for x, y in dataset:
        model.learn_one(x, y)

    assert model.height > 0



from __future__ import annotations

import math

import numpy as np
import pandas as pd
import pytest

from river import proba


@pytest.mark.parametrize(
    "p",
    [
        pytest.param(
            p,
            id=f"{p=}",
        )
        for p in [1, 3, 5]
    ],
)
def test_univariate_multivariate_consistency(p):
    X = pd.DataFrame(np.random.random((30, p)), columns=range(p))

    multi = proba.MultivariateGaussian()
    single = {c: proba.Gaussian() for c in X.columns}

    for x in X.to_dict(orient="records"):
        multi.update(x)
        for c, s in single.items():
            s.update(x[c])

    for c in X.columns:
        assert math.isclose(multi.mu[c], single[c].mu)
        assert math.isclose(multi.sigma[c][c], single[c].sigma)



from __future__ import annotations

import math

from sklearn import metrics as sk_metrics

from river import metrics


def test_log_loss():
    metric = metrics.LogLoss()

    y_true = [True, False, False, True]
    y_pred = [0.9, 0.1, 0.2, 0.65]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
        metric.update(yt, yp)

        if i >= 1:
            assert math.isclose(metric.get(), sk_metrics.log_loss(y_true[: i + 1], y_pred[: i + 1]))

    metric.revert(y_true[-1], y_pred[-1])
    assert math.isclose(metric.get(), sk_metrics.log_loss(y_true[:-1], y_pred[:-1]))



from __future__ import annotations


def test_agg_lag():
    """

    Let's say we have daily data about the number of customers and the revenue for two shops.

    >>> X = [
    ...     {'shop': '7/11', 'customers': 10, 'revenue': 420},
    ...     {'shop': 'Kmart', 'customers': 10, 'revenue': 386},
    ...     {'shop': '7/11', 'customers': 20, 'revenue': 145},
    ...     {'shop': 'Kmart', 'customers': 5, 'revenue': 87},
    ...     {'shop': '7/11', 'customers': 15, 'revenue': 276},
    ...     {'shop': 'Kmart', 'customers': 10, 'revenue': 149},
    ...     {'shop': '7/11', 'customers': 30, 'revenue': 890},
    ...     {'shop': 'Kmart', 'customers': 40, 'revenue': 782},
    ...     {'shop': '7/11', 'customers': 20, 'revenue': 403},
    ...     {'shop': 'Kmart', 'customers': 35, 'revenue': 218},
    ... ]

    At each time step, we want to use the number of customers and revenue from 2 time steps ago:

    >>> from river.feature_extraction import Agg
    >>> from river.stats import Shift

    >>> agg = (
    ...     Agg("customers", None, Shift(2)) +
    ...     Agg("revenue", None, Shift(2))
    ... )
    >>> for x in X:
    ...     agg.learn_one(x)
    ...     print(agg.transform_one(x))
    {'revenue_shift_2': None, 'customers_shift_2': None}
    {'revenue_shift_2': None, 'customers_shift_2': None}
    {'revenue_shift_2': 420, 'customers_shift_2': 10}
    {'revenue_shift_2': 386, 'customers_shift_2': 10}
    {'revenue_shift_2': 145, 'customers_shift_2': 20}
    {'revenue_shift_2': 87, 'customers_shift_2': 5}
    {'revenue_shift_2': 276, 'customers_shift_2': 15}
    {'revenue_shift_2': 149, 'customers_shift_2': 10}
    {'revenue_shift_2': 890, 'customers_shift_2': 30}
    {'revenue_shift_2': 782, 'customers_shift_2': 40}

    We can also specify multiple lags for a given feature.

    >>> from river.compose import TransformerUnion

    >>> agg = TransformerUnion(*[
    ...     Agg("customers", None, Shift(d))
    ...     for d in [1, 2, 3]
    ... ])
    >>> for x in X:
    ...     agg.learn_one(x)
    ...     print(agg.transform_one(x))
    {'customers_shift_3': None, 'customers_shift_2': None, 'customers_shift_1': None}
    {'customers_shift_3': None, 'customers_shift_2': None, 'customers_shift_1': 10}
    {'customers_shift_3': None, 'customers_shift_2': 10, 'customers_shift_1': 10}
    {'customers_shift_3': 10, 'customers_shift_2': 10, 'customers_shift_1': 20}
    {'customers_shift_3': 10, 'customers_shift_2': 20, 'customers_shift_1': 5}
    {'customers_shift_3': 20, 'customers_shift_2': 5, 'customers_shift_1': 15}
    {'customers_shift_3': 5, 'customers_shift_2': 15, 'customers_shift_1': 10}
    {'customers_shift_3': 15, 'customers_shift_2': 10, 'customers_shift_1': 30}
    {'customers_shift_3': 10, 'customers_shift_2': 30, 'customers_shift_1': 40}
    {'customers_shift_3': 30, 'customers_shift_2': 40, 'customers_shift_1': 20}

    As of now, we're looking at lagged values over all the data. It might make more sense to look
    at the lags per shop.

    >>> agg = Agg("customers", "shop", Shift(1))
    >>> for x in X:
    ...     agg.learn_one(x)
    ...     print(agg.transform_one(x))
    {'customers_shift_1_by_shop': None}
    {'customers_shift_1_by_shop': None}
    {'customers_shift_1_by_shop': 10}
    {'customers_shift_1_by_shop': 10}
    {'customers_shift_1_by_shop': 20}
    {'customers_shift_1_by_shop': 5}
    {'customers_shift_1_by_shop': 15}
    {'customers_shift_1_by_shop': 10}
    {'customers_shift_1_by_shop': 30}
    {'customers_shift_1_by_shop': 40}

    """


def test_target_agg_lag():
    """

    >>> dataset = [
    ...     ({'country': 'France'}, 42),
    ...     ({'country': 'Sweden'}, 16),
    ...     ({'country': 'France'}, 24),
    ...     ({'country': 'Sweden'}, 58),
    ...     ({'country': 'Sweden'}, 20),
    ...     ({'country': 'France'}, 50),
    ...     ({'country': 'France'}, 10),
    ...     ({'country': 'Sweden'}, 80)
    ... ]

    Let's extract the two last values of the target at each time step.

    >>> from river.feature_extraction import TargetAgg
    >>> from river.stats import Shift

    >>> agg = TargetAgg(None, Shift(1)) + TargetAgg(None, Shift(2))
    >>> for x, y in dataset:
    ...     print(agg.transform_one(x))
    ...     agg.learn_one(x, y)
    {'y_shift_2': None, 'y_shift_1': None}
    {'y_shift_2': None, 'y_shift_1': None}
    {'y_shift_2': None, 'y_shift_1': 42}
    {'y_shift_2': 42, 'y_shift_1': 16}
    {'y_shift_2': 16, 'y_shift_1': 24}
    {'y_shift_2': 24, 'y_shift_1': 58}
    {'y_shift_2': 58, 'y_shift_1': 20}
    {'y_shift_2': 20, 'y_shift_1': 50}

    We can also calculate the lags with different groups:

    >>> agg = TargetAgg("country", Shift(1)) + TargetAgg("country", Shift(2))
    >>> for x, y in dataset:
    ...     print(agg.transform_one(x))
    ...     agg.learn_one(x, y)
    {'y_shift_2_by_country': None, 'y_shift_1_by_country': None}
    {'y_shift_2_by_country': None, 'y_shift_1_by_country': None}
    {'y_shift_2_by_country': None, 'y_shift_1_by_country': None}
    {'y_shift_2_by_country': None, 'y_shift_1_by_country': None}
    {'y_shift_2_by_country': None, 'y_shift_1_by_country': 16}
    {'y_shift_2_by_country': None, 'y_shift_1_by_country': 42}
    {'y_shift_2_by_country': 42, 'y_shift_1_by_country': 24}
    {'y_shift_2_by_country': 16, 'y_shift_1_by_country': 58}

    """



from __future__ import annotations

import importlib
import inspect
import math
import random
import typing

import numpy as np
import pytest

from river import optim, utils


def losses() -> typing.Iterable[optim.losses.Loss]:
    for _, loss in inspect.getmembers(
        importlib.import_module("river.optim.losses"),
        lambda x: inspect.isclass(x)
        and not inspect.isabstract(x)
        and not issubclass(x, optim.losses.CrossEntropy),
    ):
        yield loss()


@pytest.mark.parametrize(
    "loss",
    [pytest.param(loss, id=loss.__class__.__name__) for loss in losses()],
)
def test_loss_batch_online_equivalence(loss):
    y_true = np.random.randint(2, size=30)
    y_pred = np.random.uniform(-10, 10, size=30)

    for yt, yp, g in zip(y_true, y_pred, loss.gradient(y_true, y_pred)):
        assert math.isclose(loss.gradient(yt, yp), g, abs_tol=1e-9)


def optimizers() -> typing.Iterable[optim.base.Optimizer]:
    for _, optimizer in inspect.getmembers(
        importlib.import_module("river.optim"),
        lambda x: inspect.isclass(x)
        and issubclass(x, optim.base.Optimizer)
        and x is not optim.base.Optimizer,
    ):
        for params in optimizer._unit_test_params():
            yield optimizer(**params)


@pytest.mark.parametrize(
    "optimizer",
    [pytest.param(optimizer, id=optimizer.__class__.__name__) for optimizer in optimizers()],
)
def test_optimizer_step_with_dict_same_as_step_with_vector_dict(optimizer):
    w_dict = {i: random.uniform(-5, 5) for i in range(10)}
    w_vector = utils.VectorDict(w_dict)

    g_dict = {i: random.uniform(-5, 5) for i in range(10)}
    g_vector = utils.VectorDict(g_dict)

    w_dict = optimizer._step_with_dict(w_dict, g_dict)
    try:
        w_vector = optimizer.clone()._step_with_vector(w_vector, g_vector)
    except NotImplementedError:
        pytest.skip("step_with_vector not implemented")

    for i, w in w_vector.to_dict().items():
        assert math.isclose(w, w_dict[i])


@pytest.mark.parametrize(
    "optimizer",
    [pytest.param(optimizer, id=optimizer.__class__.__name__) for optimizer in optimizers()],
)
def test_optimizer_step_with_dict_same_as_step_with_numpy_array(optimizer):
    w_dict = {i: random.uniform(-5, 5) for i in range(10)}
    w_vector = np.asarray(list(w_dict.values()))

    g_dict = {i: random.uniform(-5, 5) for i in range(10)}
    g_vector = np.asarray(list(g_dict.values()))

    w_dict = optimizer._step_with_dict(w_dict, g_dict)
    try:
        w_vector = optimizer.clone()._step_with_vector(w_vector, g_vector)
    except NotImplementedError:
        pytest.skip("step_with_vector not implemented")

    for i, w in dict(enumerate(w_vector)).items():
        assert math.isclose(w, w_dict[i])



from __future__ import annotations

import numpy as np
import pytest

from river import drift

np.random.seed(12345)
data_stream_1 = np.concatenate((np.random.randint(2, size=1000), np.random.randint(8, size=1000)))

np.random.seed(12345)
data_stream_2 = np.concatenate(
    [
        [np.random.binomial(1, 0.2) for _ in range(1000)],
        [np.random.binomial(1, 0.8) for _ in range(1000)],
    ]
).astype(int)

np.random.seed(12345)
data_stream_3 = np.concatenate(
    (
        np.random.normal(0.0, 0.1, 500) > 0,
        np.random.normal(0.25, 0.1, 500) > 0,
        np.random.normal(0.0, 0.1, 500) > 0,
        np.random.normal(0.25, 0.1, 500) > 0,
    )
).astype(int)


def test_adwin():
    expected_indices = [1023]
    detected_indices = perform_test(drift.ADWIN(), data_stream_1)

    assert detected_indices == expected_indices


def test_ddm():
    expected_indices = [1049]
    detected_indices = perform_test(drift.binary.DDM(), data_stream_2)
    assert detected_indices == expected_indices

    expected_indices = []
    detected_indices = perform_test(drift.binary.DDM(), np.ones(1000))
    assert detected_indices == expected_indices


def test_eddm():
    expected_indices = [1059]
    detected_indices = perform_test(drift.binary.EDDM(alpha=0.9, beta=0.8), data_stream_2)
    assert detected_indices == expected_indices


def test_hddm_a():
    hddm_a = drift.binary.HDDM_A()
    expected_indices = [1047]
    detected_indices = perform_test(hddm_a, data_stream_2)
    assert detected_indices == expected_indices

    # Second test, more abrupt drifts
    hddm_a = drift.binary.HDDM_A(two_sided_test=True)
    expected_indices = [531, 1015, 1545]
    detected_indices = perform_test(hddm_a, data_stream_3)
    assert detected_indices == expected_indices


def test_hddm_w():
    hddm_w = drift.binary.HDDM_W()
    expected_indices = [1018]
    detected_indices = perform_test(hddm_w, data_stream_2)
    assert detected_indices == expected_indices

    # Second test, more abrupt drifts
    hddm_w = drift.binary.HDDM_W(two_sided_test=True)
    expected_indices = [507, 1032, 1508]
    detected_indices = perform_test(hddm_w, data_stream_3)
    assert detected_indices == expected_indices


def test_kswin():
    kswin = drift.KSWIN(alpha=0.0001, window_size=200, stat_size=100, seed=42)
    expected_indices = [1042]
    detected_indices = perform_test(kswin, data_stream_1)
    assert detected_indices == expected_indices


def test_kswin_coverage():
    with pytest.raises(ValueError):
        drift.KSWIN(alpha=-0.1)

    with pytest.raises(ValueError):
        drift.KSWIN(alpha=1.1)

    try:
        drift.KSWIN(window_size=-10)
    except ValueError:
        assert True
    else:
        assert False
    try:
        drift.KSWIN(window_size=10, stat_size=30)
    except ValueError:
        assert True
    else:
        assert False


def test_page_hinkley():
    expected_indices = [588, 1681]
    detected_indices = perform_test(drift.PageHinkley(mode="up"), data_stream_3)

    assert detected_indices == expected_indices

    expected_indices = [1172]
    detected_indices = perform_test(drift.PageHinkley(mode="down"), data_stream_3)

    assert detected_indices == expected_indices

    expected_indices = [588, 1097, 1585]
    detected_indices = perform_test(drift.PageHinkley(mode="both"), data_stream_3)

    assert detected_indices == expected_indices


def perform_test(drift_detector, data_stream):
    detected_indices = []
    for i, val in enumerate(data_stream):
        drift_detector.update(val)
        if drift_detector.drift_detected:
            detected_indices.append(i)
    return detected_indices



from __future__ import annotations

import pandas as pd

from river import datasets, linear_model, metrics, multiclass, preprocessing, stream


def test_online_batch_consistent():
    # Batch

    batch = preprocessing.StandardScaler() | multiclass.OneVsRestClassifier(
        linear_model.LogisticRegression()
    )

    dataset = datasets.ImageSegments()

    batch_metric = metrics.MacroF1()

    for i, x in enumerate(pd.read_csv(dataset.path, chunksize=1)):
        y = x.pop("category")
        y_pred = batch.predict_many(x)
        batch.learn_many(x, y)

        for yt, yp in zip(y, y_pred):
            if yp is not None:
                batch_metric.update(yt, yp)

        if i == 30:
            break

    # Online

    online = preprocessing.StandardScaler() | multiclass.OneVsRestClassifier(
        linear_model.LogisticRegression()
    )

    online_metric = metrics.MacroF1()

    X = pd.read_csv(dataset.path)
    Y = X.pop("category")

    for i, (x, y) in enumerate(stream.iter_pandas(X, Y)):
        y_pred = online.predict_one(x)
        online.learn_one(x, y)

        if y_pred is not None:
            online_metric.update(y, y_pred)

        if i == 30:
            break

    assert online_metric.get() == batch_metric.get()



from __future__ import annotations


def test_missing_features():
    """Checks that HalfSpaceTrees works even if a feature is missing.

    >>> import random
    >>> from river import anomaly
    >>> from river import compose
    >>> from river import datasets
    >>> from river import metrics
    >>> from river import preprocessing

    >>> model = compose.Pipeline(
    ...     preprocessing.MinMaxScaler(),
    ...     anomaly.HalfSpaceTrees(seed=42)
    ... )

    >>> auc = metrics.ROCAUC()

    >>> features = list(next(iter(datasets.CreditCard()))[0].keys())
    >>> random.seed(42)

    >>> for x, y in datasets.CreditCard().take(8000):
    ...     del x[random.choice(features)]
    ...     score = model.score_one(x)
    ...     model.learn_one(x, y)
    ...     auc.update(y, score)

    >>> auc
    ROCAUC: 88.68%

    """



from __future__ import annotations

import collections
import math

from sklearn import metrics as sk_metrics

from river import metrics, utils


def test_multi_fbeta():
    fbeta = metrics.MultiFBeta(betas={0: 0.25, 1: 1, 2: 4}, weights={0: 1, 1: 1, 2: 2})
    sk_fbeta = sk_metrics.fbeta_score
    y_true = [0, 1, 2, 2, 2]
    y_pred = [0, 1, 0, 2, 1]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
        fbeta.update(yt, yp)

        if i >= 2:
            fbeta_0, _, _ = sk_fbeta(y_true[: i + 1], y_pred[: i + 1], beta=0.25, average=None)
            _, fbeta_1, _ = sk_fbeta(y_true[: i + 1], y_pred[: i + 1], beta=1, average=None)
            _, _, fbeta_2 = sk_fbeta(y_true[: i + 1], y_pred[: i + 1], beta=4, average=None)

            multi_fbeta = fbeta_0 * 1 + fbeta_1 * 1 + fbeta_2 * 2
            multi_fbeta /= 1 + 1 + 2

            assert math.isclose(fbeta.get(), multi_fbeta)


def test_rolling_multi_fbeta():
    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    fbeta = utils.Rolling(
        metrics.MultiFBeta(betas={0: 0.25, 1: 1, 2: 4}, weights={0: 1, 1: 1, 2: 2}),
        window_size=3,
    )
    n = fbeta.window_size
    sk_fbeta = sk_metrics.fbeta_score
    y_true = [0, 1, 2, 2, 2]
    y_pred = [0, 1, 0, 2, 1]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
        fbeta.update(yt, yp)

        if i >= 2:
            sk_y_true, sk_y_pred = tail(y_true[: i + 1], n), tail(y_pred[: i + 1], n)
            fbeta_0, _, _ = sk_fbeta(sk_y_true, sk_y_pred, beta=0.25, average=None)
            _, fbeta_1, _ = sk_fbeta(sk_y_true, sk_y_pred, beta=1, average=None)
            _, _, fbeta_2 = sk_fbeta(sk_y_true, sk_y_pred, beta=4, average=None)

            multi_fbeta = fbeta_0 * 1 + fbeta_1 * 1 + fbeta_2 * 2
            multi_fbeta /= 1 + 1 + 2

            assert math.isclose(fbeta.get(), multi_fbeta)



from __future__ import annotations

import importlib
import inspect

import pytest

from river import (
    bandit,
    datasets,
    evaluate,
    linear_model,
    metrics,
    model_selection,
    optim,
    preprocessing,
)


def test_1259():
    """

    https://github.com/online-ml/river/issues/1259

    >>> from river import bandit
    >>> from river import datasets
    >>> from river import evaluate
    >>> from river import linear_model
    >>> from river import metrics
    >>> from river import model_selection
    >>> from river import optim
    >>> from river import preprocessing

    >>> models = [
    ...     linear_model.LogisticRegression(optimizer=optim.SGD(lr=lr))
    ...     for lr in [0.0001, 0.001, 1e-05, 0.01]
    ... ]

    >>> dataset = datasets.Phishing()
    >>> model = (
    ...     preprocessing.StandardScaler() |
    ...     model_selection.BanditClassifier(
    ...         models,
    ...         metric=metrics.Accuracy(),
    ...         policy=bandit.Exp3(
    ...             gamma=0.5,
    ...             seed=42
    ...         )
    ...     )
    ... )
    >>> metric = metrics.Accuracy()

    >>> evaluate.progressive_val_score(dataset, model, metric)
    Accuracy: 87.20%

    """


@pytest.mark.parametrize(
    "policy",
    [
        pytest.param(
            policy(**params),
            id=f"{policy.__name__}",
        )
        for _, policy in inspect.getmembers(
            importlib.import_module("river.bandit"),
            lambda obj: inspect.isclass(obj)
            and issubclass(obj, bandit.base.Policy)
            and not issubclass(obj, bandit.base.ContextualPolicy)
            and obj.__name__ not in {"ThompsonSampling"},
        )
        for params in policy._unit_test_params()
    ],
)
def test_bandit_classifier_with_each_policy(policy):
    models = [
        linear_model.LogisticRegression(optimizer=optim.SGD(lr=lr))
        for lr in [0.0001, 0.001, 1e-05, 0.01]
    ]

    dataset = datasets.Phishing()
    model = preprocessing.StandardScaler() | model_selection.BanditClassifier(
        models, metric=metrics.Accuracy(), policy=policy
    )
    metric = metrics.Accuracy()

    score = evaluate.progressive_val_score(dataset, model, metric)
    assert score.get() > 0.5


@pytest.mark.parametrize(
    "policy",
    [
        pytest.param(
            policy(**params),
            id=f"{policy.__name__}",
        )
        for _, policy in inspect.getmembers(
            importlib.import_module("river.bandit"),
            lambda obj: inspect.isclass(obj)
            and issubclass(obj, bandit.base.Policy)
            and not issubclass(obj, bandit.base.ContextualPolicy)
            and obj.__name__ not in {"ThompsonSampling", "Exp3"},
        )
        for params in policy._unit_test_params()
    ],
)
def test_bandit_regressor_with_each_policy(policy):
    models = [
        linear_model.LinearRegression(optimizer=optim.SGD(lr=lr))
        for lr in [0.0001, 0.001, 1e-05, 0.01]
    ]

    dataset = datasets.TrumpApproval()
    model = preprocessing.StandardScaler() | model_selection.BanditRegressor(
        models, metric=metrics.MSE(), policy=policy
    )
    metric = metrics.MSE()

    score = evaluate.progressive_val_score(dataset, model, metric)
    assert score.get() < 300



from __future__ import annotations

import numpy as np
import pytest

from river.utils import VectorDict


def test_vectordict():
    # test empty init
    x = dict()
    vx = VectorDict()
    assert vx == x

    # test basics
    x = {"a": 8, "b": -1.2, 4: 2.7}
    vx = VectorDict(x)
    assert vx == x
    assert vx["a"] == 8
    assert vx[4] == 2.7
    vx[9] = 8.9
    assert x[9] == vx[9] == 8.9

    # test copy
    x = {"a": 8, "b": -1.2, 4: 2.7}
    vx = VectorDict(x, copy=True)
    assert vx == x
    vx["a"] = 2
    assert x["a"] == 8
    assert vx["a"] == 2

    # test operations
    x = {"a": 1, "b": -5, "c": -3}
    y = {"a": 2, "b": 0.5, "d": 4}
    vx = VectorDict(x)
    vy = VectorDict(y)
    assert vx == vx == x
    assert +vx == vx == x
    assert -vx == {"a": -1, "b": 5, "c": 3}
    assert vx + 2 == 2 + vx == {"a": 3, "b": -3, "c": -1}
    assert vx * 2 == 2 * vx == {"a": 2, "b": -10, "c": -6}
    assert vx - 2 == {"a": -1, "b": -7, "c": -5}
    assert 2 - vx == -(vx - 2)
    assert vx / 2 == {"a": 0.5, "b": -2.5, "c": -1.5}
    assert 2 / vx == {"a": 2, "b": -0.4, "c": -2 / 3}
    assert vx + vy == vy + vx == {"a": 3, "b": -4.5, "c": -3, "d": 4}
    assert vx - vy == {"a": -1, "b": -5.5, "c": -3, "d": -4}
    assert vx**2 == pow(vx, 2) == {"a": 1, "b": 25, "c": 9}
    assert vx * vy == vy * vx == {"a": 2, "b": -2.5, "d": 0, "c": 0}
    assert vx / vx == {"a": 1.0, "b": 1.0, "c": 1.0}
    with pytest.raises(ZeroDivisionError):
        vx / vy
    assert vx @ vy == vy @ vx == -0.5
    vz = VectorDict(x, copy=True)
    vz += 2
    assert vz == vx + 2
    vz = VectorDict(x, copy=True)
    vz -= 2
    assert vz == vx - 2
    vz = VectorDict(x, copy=True)
    vz *= 2
    assert vz == vx * 2
    vz = VectorDict(x, copy=True)
    vz /= 2
    assert vz == vx / 2
    vz = VectorDict(x, copy=True)
    vz += vy
    assert vz == vx + vy
    vz = VectorDict(x, copy=True)
    vz -= vy
    assert vz == vx - vy
    vz = VectorDict(x, copy=True)
    vz *= vy
    assert vz == vx * vy
    vz = VectorDict(x, copy=True)
    vz /= vz
    assert vz == vx / vx
    vz = VectorDict(x, copy=True)
    vz **= 2
    assert vz == vx**2

    # test default_factory
    x = {"a": 1, "b": -5}
    y = {"b": 0.5, "d": 4, "e": 3, "f": 8}
    counter = iter(range(100))
    vx = VectorDict(x, default_factory=counter.__next__)
    vy = VectorDict(y)
    assert vx @ vy == 16.5
    assert counter.__next__() == 3
    assert x["f"] == 2

    # test mask
    x = {"a": 1, "b": -5, "e": 2}
    y = {"b": 0.5, "d": 4, "e": 3, "f": 8}
    z = {"b": 4, "d": 2, "g": -1}
    vx = VectorDict(x)
    vy = VectorDict(y)
    assert vx + vy == vy + vx == {"a": 1, "b": -4.5, "d": 4, "e": 5, "f": 8}
    vy = VectorDict(y, mask=z)
    assert vx + vy == vy + vx == {"a": 1, "b": -4.5, "d": 4, "e": 2}
    vy = VectorDict(y).with_mask(z.keys())
    assert vx + vy == vy + vx == {"a": 1, "b": -4.5, "d": 4, "e": 2}
    vy = VectorDict(y).with_mask(x)
    assert vy / vx == {"b": -0.1, "a": 0.0, "e": 1.5}

    # test export
    x = {"a": 1, "b": -5}
    vx = VectorDict(x)
    nx = vx.to_numpy(["b", "c"])
    assert isinstance(nx, np.ndarray)
    assert (vx.to_numpy(["b", "c"]) == np.array([-5, 0])).all()

    # other methods
    x = {"a": 1, "b": -5}
    vx = VectorDict(x)
    assert vx.abs() == abs(vx) == {"a": 1, "b": 5}
    assert vx.min() == -5
    assert vx.max() == 1
    assert vx.with_mask(["a"]).min() == 1
    assert vx.with_mask(["b"]).max() == -5
    assert vx.minimum(-2) == {"a": -2, "b": -5}
    assert vx.maximum(-2) == {"a": 1, "b": -2}
    y = {"b": 0.5, "c": 4}
    vy = VectorDict(y)
    assert vx.minimum(vy) == vy.minimum(vx) == {"a": 0, "b": -5, "c": 0}
    assert vx.maximum(vy) == vy.maximum(vx) == {"a": 1, "b": 0.5, "c": 4}



from __future__ import annotations

import collections
import copy
import functools
import importlib
import inspect
import pickle
import platform
import random

import numpy as np
import pytest
from sklearn import metrics as sk_metrics

from river import metrics, utils


def load_metrics():
    """Yields all the metrics."""

    for name, obj in inspect.getmembers(importlib.import_module("river.metrics"), inspect.isclass):
        if name == "Metrics":
            continue

        if inspect.isabstract(obj):
            continue

        elif name == "RegressionMultiOutput":
            yield obj(metric=metrics.MSE())
            continue

        try:
            sig = inspect.signature(obj)
            yield obj(
                **{
                    param.name: param.default if param.default != param.empty else 5
                    for param in sig.parameters.values()
                }
            )
        except ValueError:
            yield obj()


@pytest.mark.parametrize(
    "metric",
    [pytest.param(metric, id=type(metric).__name__) for metric in load_metrics()],
)
def test_pickling(metric):
    assert isinstance(pickle.loads(pickle.dumps(metric)), metric.__class__)
    assert isinstance(copy.deepcopy(metric), metric.__class__)


@pytest.mark.parametrize(
    "metric",
    [pytest.param(metric, id=type(metric).__name__) for metric in load_metrics()],
)
def test_repr(metric):
    repr(metric)

    for y_true, y_pred, _ in generate_test_cases(metric=metric, n=10):
        m = copy.deepcopy(metric)

        for yt, yp in zip(y_true, y_pred):
            m.update(yt, yp)
            repr(metric)


def generate_test_cases(metric, n):
    """Yields n (y_true, y_pred, sample_weight) triplets for a given metric."""

    sample_weights = [random.random() for _ in range(n)]

    if isinstance(metric, metrics.base.ClassificationMetric):
        y_true = [random.choice([False, True]) for _ in range(n)]
        if metric.requires_labels:
            y_pred = [random.choice([False, True]) for _ in range(n)]
        else:
            y_pred = [dict(zip([False, True], np.random.dirichlet(np.ones(2)))) for _ in range(n)]
        yield y_true, y_pred, sample_weights

    if isinstance(metric, metrics.base.MultiClassMetric):
        y_true = [random.choice([0, 1, 2]) for _ in range(n)]
        if metric.requires_labels:
            y_pred = [random.choice([0, 1, 2]) for _ in range(n)]
        else:
            y_pred = [dict(zip([0, 1, 2], np.random.dirichlet(np.ones(3)))) for _ in range(n)]
        yield y_true, y_pred, sample_weights

    if isinstance(metric, metrics.base.RegressionMetric):
        yield (
            [random.random() for _ in range(n)],
            [random.random() for _ in range(n)],
            sample_weights,
        )


def partial(f, **kwargs):
    return functools.update_wrapper(functools.partial(f, **kwargs), f)


def roc_auc_score(y_true, y_score):
    """
    This functions is a wrapper to the scikit-learn roc_auc_score function.
    It was created because the scikit version utilizes array of scores and
    may raise a ValueError if there is only one class present in y_true.
    This wrapper returns 0 if y_true has only one class and
    deals with the scores.
    """
    nonzero = np.count_nonzero(y_true)
    if nonzero == 0 or nonzero == len(y_true):
        return 0

    scores = [s[True] for s in y_score]

    return sk_metrics.roc_auc_score(y_true, scores)


TEST_CASES = [
    (metrics.Accuracy(), sk_metrics.accuracy_score),
    (metrics.Precision(), partial(sk_metrics.precision_score, zero_division=0)),
    (
        metrics.MacroPrecision(),
        partial(sk_metrics.precision_score, average="macro", zero_division=0),
    ),
    (
        metrics.MicroPrecision(),
        partial(sk_metrics.precision_score, average="micro", zero_division=0),
    ),
    (
        metrics.WeightedPrecision(),
        partial(sk_metrics.precision_score, average="weighted", zero_division=0),
    ),
    (metrics.Recall(), partial(sk_metrics.recall_score, zero_division=0)),
    (
        metrics.MacroRecall(),
        partial(sk_metrics.recall_score, average="macro", zero_division=0),
    ),
    (
        metrics.MicroRecall(),
        partial(sk_metrics.recall_score, average="micro", zero_division=0),
    ),
    (
        metrics.WeightedRecall(),
        partial(sk_metrics.recall_score, average="weighted", zero_division=0),
    ),
    (
        metrics.FBeta(beta=0.5),
        partial(sk_metrics.fbeta_score, beta=0.5, zero_division=0),
    ),
    (
        metrics.MacroFBeta(beta=0.5),
        partial(sk_metrics.fbeta_score, beta=0.5, average="macro", zero_division=0),
    ),
    (
        metrics.MicroFBeta(beta=0.5),
        partial(sk_metrics.fbeta_score, beta=0.5, average="micro", zero_division=0),
    ),
    (
        metrics.WeightedFBeta(beta=0.5),
        partial(sk_metrics.fbeta_score, beta=0.5, average="weighted", zero_division=0),
    ),
    (metrics.F1(), partial(sk_metrics.f1_score, zero_division=0)),
    (metrics.MacroF1(), partial(sk_metrics.f1_score, average="macro", zero_division=0)),
    (metrics.MicroF1(), partial(sk_metrics.f1_score, average="micro", zero_division=0)),
    (
        metrics.WeightedF1(),
        partial(sk_metrics.f1_score, average="weighted", zero_division=0),
    ),
    (metrics.MCC(), sk_metrics.matthews_corrcoef),
    (metrics.MAE(), sk_metrics.mean_absolute_error),
    (metrics.MSE(), sk_metrics.mean_squared_error),
    (metrics.Homogeneity(), sk_metrics.homogeneity_score),
    (metrics.Completeness(), sk_metrics.completeness_score),
    (metrics.VBeta(beta=0.5), partial(sk_metrics.v_measure_score, beta=0.5)),
    (metrics.FowlkesMallows(), sk_metrics.fowlkes_mallows_score),
    (metrics.Rand(), sk_metrics.rand_score),
    (metrics.AdjustedRand(), sk_metrics.adjusted_rand_score),
    (metrics.MutualInfo(), sk_metrics.mutual_info_score),
    (
        metrics.NormalizedMutualInfo(average_method="min"),
        partial(sk_metrics.normalized_mutual_info_score, average_method="min"),
    ),
    (
        metrics.NormalizedMutualInfo(average_method="max"),
        partial(sk_metrics.normalized_mutual_info_score, average_method="max"),
    ),
    (
        metrics.NormalizedMutualInfo(average_method="arithmetic"),
        partial(sk_metrics.normalized_mutual_info_score, average_method="arithmetic"),
    ),
    (
        metrics.NormalizedMutualInfo(average_method="geometric"),
        partial(sk_metrics.normalized_mutual_info_score, average_method="geometric"),
    ),
    (
        metrics.AdjustedMutualInfo(average_method="max"),
        partial(sk_metrics.adjusted_mutual_info_score, average_method="max"),
    ),
    (
        metrics.AdjustedMutualInfo(average_method="arithmetic"),
        partial(sk_metrics.adjusted_mutual_info_score, average_method="arithmetic"),
    ),
    (
        metrics.AdjustedMutualInfo(average_method="geometric"),
        partial(sk_metrics.adjusted_mutual_info_score, average_method="geometric"),
    ),
    (metrics.Jaccard(), partial(sk_metrics.jaccard_score, average="binary")),
    (metrics.MacroJaccard(), partial(sk_metrics.jaccard_score, average="macro")),
    (metrics.MicroJaccard(), partial(sk_metrics.jaccard_score, average="micro")),
    (metrics.WeightedJaccard(), partial(sk_metrics.jaccard_score, average="weighted")),
    (metrics.RollingROCAUC(), roc_auc_score),
]

# HACK: not sure why this is needed, see this CI run https://github.com/online-ml/river/runs/7992357532?check_suite_focus=true
if platform.system() != "Linux":
    TEST_CASES.append(
        (
            metrics.AdjustedMutualInfo(average_method="min"),
            partial(sk_metrics.adjusted_mutual_info_score, average_method="min"),
        )
    )


@pytest.mark.parametrize(
    "metric, sk_metric",
    [
        pytest.param(metric, sk_metric, id=f"{metric.__class__.__name__}")
        for metric, sk_metric in TEST_CASES
    ],
)
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.filterwarnings("ignore::sklearn.exceptions.UndefinedMetricWarning")
def test_metric(metric, sk_metric):
    for y_true, y_pred, sample_weights in generate_test_cases(metric=metric, n=30):
        m = copy.deepcopy(metric)
        for i, (yt, yp, w) in enumerate(zip(y_true, y_pred, sample_weights)):
            if metric.works_with_weights:
                m.update(y_true=yt, y_pred=yp, w=w)
            else:
                m.update(y_true=yt, y_pred=yp)

            if i >= 1:
                if metric.works_with_weights:
                    kwargs = {"sample_weight": sample_weights[: i + 1]}
                else:
                    kwargs = {}
                assert abs(m.get() - sk_metric(y_true[: i + 1], y_pred[: i + 1], **kwargs)) < 1e-6


@pytest.mark.parametrize(
    "metric, sk_metric",
    [
        pytest.param(metric, sk_metric, id=f"{metric.__class__.__name__}")
        for metric, sk_metric in TEST_CASES
    ],
)
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.filterwarnings("ignore::sklearn.exceptions.UndefinedMetricWarning")
def test_rolling_metric(metric, sk_metric):
    def tail(iterable, n):
        return collections.deque(iterable, maxlen=n)

    for n in (1, 2, 5, 10):
        for y_true, y_pred, _ in generate_test_cases(metric=metric, n=30):
            m = utils.Rolling(copy.deepcopy(metric), window_size=n)

            # Check str works
            str(m)

            for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
                m.update(y_true=yt, y_pred=yp)

                if i >= 1:
                    assert (
                        abs(
                            m.get()
                            - sk_metric(
                                tail(y_true[: i + 1], n),
                                tail(y_pred[: i + 1], n),
                            )
                        )
                        < 1e-10
                    )


def test_compose():
    metrics.MAE() + metrics.MSE()
    metrics.Accuracy() + metrics.LogLoss()

    with pytest.raises(ValueError):
        _ = metrics.MSE() + metrics.LogLoss()

    with pytest.raises(ValueError):
        _ = metrics.MSE() + metrics.MAE() + metrics.LogLoss()



from __future__ import annotations

import math

import numpy as np
import pandas as pd

from river import datasets, preprocessing, stream


def test_standard_scaler_one_many_consistent():
    """Checks that using learn_one or learn_many produces the same result."""

    for with_std in (False, True):
        X = pd.read_csv(datasets.TrumpApproval().path)

        one = preprocessing.StandardScaler(with_std=with_std)
        for x, _ in stream.iter_pandas(X):
            one.learn_one(x)

        many = preprocessing.StandardScaler(with_std=with_std)
        for xb in np.array_split(X, 10):
            many.learn_many(xb)

        for i in X:
            assert math.isclose(one.counts[i], many.counts[i])
            assert math.isclose(one.means[i], many.means[i])
            assert math.isclose(one.vars[i], many.vars[i])


def test_standard_scaler_shuffle_columns():
    """Checks that learn_many works identically whether columns are shuffled or not."""

    X = pd.read_csv(datasets.TrumpApproval().path)

    normal = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        normal.learn_many(xb)

    shuffled = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        cols = np.random.permutation(X.columns)
        shuffled.learn_many(xb[cols])

    for i in X:
        assert math.isclose(shuffled.counts[i], shuffled.counts[i])
        assert math.isclose(shuffled.means[i], shuffled.means[i])
        assert math.isclose(shuffled.vars[i], shuffled.vars[i])


def test_standard_scaler_add_remove_columns():
    """Checks that no exceptions are raised whenever columns are dropped and/or added."""

    X = pd.read_csv(datasets.TrumpApproval().path)

    ss = preprocessing.StandardScaler()
    for xb in np.array_split(X, 10):
        # Pick half of the columns at random
        cols = np.random.choice(X.columns, len(X.columns) // 2, replace=False)
        ss.learn_many(xb[cols])


def test_issue_1313():
    """

    >>> import numpy as np
    >>> import pandas as pd
    >>> from sklearn import datasets
    >>> from river import preprocessing
    >>> from river.compose import Select

    >>> X, y = datasets.make_regression(n_samples=6, n_features=2)
    >>> X = pd.DataFrame(X)
    >>> X.columns = ['feat_1','feat_2']

    >>> model = Select('feat_1') | preprocessing.StandardScaler()
    >>> X = X.astype('float32')
    >>> X.dtypes
    feat_1    float32
    feat_2    float32
    dtype: object

    >>> model.learn_many(X)
    >>> X1 = model.transform_many(X)
    >>> X1.dtypes
    feat_1    float32
    dtype: object

    """



from __future__ import annotations

from collections import deque

import numpy as np
from scipy.stats import ks_2samp

from river import stats


def test_incremental_ks_statistics():
    initial_a = np.random.normal(loc=0, scale=1, size=500)
    initial_b = np.random.normal(loc=1, scale=1, size=500)

    stream_a = np.random.normal(loc=0, scale=1, size=5000)
    stream_b = np.random.normal(loc=1, scale=1, size=5000)

    incremental_ks_statistics = []
    incremental_ks = stats.KolmogorovSmirnov(statistic="ks")
    sliding_a = deque(initial_a)
    sliding_b = deque(initial_b)

    for a, b in zip(initial_a, initial_b):
        incremental_ks.update(a, b)
    for a, b in zip(stream_a, stream_b):
        incremental_ks.revert(sliding_a.popleft(), sliding_b.popleft())
        sliding_a.append(a)
        sliding_b.append(b)
        incremental_ks.update(a, b)
        incremental_ks_statistics.append(incremental_ks.get())

    ks_2samp_statistics = []
    sliding_a = deque(initial_a)
    sliding_b = deque(initial_b)

    for a, b in zip(stream_a, stream_b):
        sliding_a.popleft()
        sliding_b.popleft()
        sliding_a.append(a)
        sliding_b.append(b)
        ks_2samp_statistics.append(ks_2samp(sliding_a, sliding_b).statistic)

    assert np.allclose(np.array(incremental_ks_statistics), np.array(ks_2samp_statistics))

    assert incremental_ks._test_ks_threshold(ca=incremental_ks._ca(p_value=0.05)) is True



from __future__ import annotations

import io

from river import stream


def test_iter_csv_custom_converter():
    example = io.StringIO("col1,col2,col3\n,1,2\n5,,4\n3,1,")

    def int_or_none(s):
        try:
            return int(s)
        except ValueError:
            return None

    params = {"converters": {"col1": int_or_none, "col2": int_or_none, "col3": int_or_none}}
    dataset = stream.iter_csv(example, **params)
    assert list(dataset) == [
        ({"col1": None, "col2": 1, "col3": 2}, None),
        ({"col1": 5, "col2": None, "col3": 4}, None),
        ({"col1": 3, "col2": 1, "col3": None}, None),
    ]


def test_iter_csv_drop_nones():
    example = io.StringIO("col1,col2,col3\n,1,2\n5,,4\n3,1,")

    def int_or_none(s):
        try:
            return int(s)
        except ValueError:
            return None

    params = {
        "converters": {"col1": int_or_none, "col2": int_or_none, "col3": int_or_none},
        "drop_nones": True,
    }
    dataset = stream.iter_csv(example, **params)
    assert list(dataset) == [
        ({"col2": 1, "col3": 2}, None),
        ({"col1": 5, "col3": 4}, None),
        ({"col1": 3, "col2": 1}, None),
    ]



from __future__ import annotations

import copy
import math
import random

import numpy as np
import pytest

from river import stats


def test_add_mean():
    A = stats.Mean()
    B = stats.Mean()
    C = stats.Mean()

    X = [random.random() for _ in range(30)]
    Y = [random.random() for _ in range(30)]
    W = [random.random() for _ in range(30)]

    for i, (x, y, w) in enumerate(zip(X, Y, W)):
        A.update(x, w)
        B.update(y, w)
        C.update(x, w)
        C.update(y, w)

        D = A + B
        assert math.isclose(C.get(), D.get())

        # TODO: clone should work instead of deepcopy
        E = copy.deepcopy(A)
        E += B
        assert math.isclose(C.get(), E.get())

        assert math.isclose(
            C.get(), np.average(X[: i + 1] + Y[: i + 1], weights=W[: i + 1] + W[: i + 1])
        )


def _weighted_var(X, W, ddof):
    average = np.average(X, weights=W)
    return np.average((X - average) ** 2, weights=W) * len(X) / (len(X) - ddof)


@pytest.mark.parametrize("ddof", [pytest.param(ddof, id=f"{ddof=}") for ddof in [0, 1, 2]])
def test_add_var(ddof):
    A = stats.Var(ddof)
    B = stats.Var(ddof)
    C = stats.Var(ddof)

    X = [random.random() for _ in range(30)]
    Y = [random.random() for _ in range(30)]
    W = [1 for _ in range(30)]

    for i, (x, y, w) in enumerate(zip(X, Y, W)):
        A.update(x, w)
        B.update(y, w)
        C.update(x, w)
        C.update(y, w)

        D = A + B
        assert math.isclose(C.get(), D.get())

        E = A.clone(include_attributes=True)
        E += B
        assert math.isclose(C.get(), E.get())

        if i >= ddof:
            assert math.isclose(
                C.get(), _weighted_var(X[: i + 1] + Y[: i + 1], W[: i + 1] + W[: i + 1], ddof=ddof)
            )


@pytest.mark.parametrize(
    "stat",
    [
        pytest.param(stat, id=stat.__class__.__name__)
        for stat in [stats.Mean(), stats.Var(ddof=0), stats.Var(ddof=1), stats.Var(ddof=2)]
    ],
)
def test_sub(stat):
    A = copy.deepcopy(stat)
    B = copy.deepcopy(stat)
    C = copy.deepcopy(stat)

    X = [random.random() for _ in range(30)]
    Y = [random.random() for _ in range(30)]
    W = [random.random() for _ in range(30)]

    for x, y, w in zip(X, Y, W):
        A.update(x, w)
        B.update(y, w)
        C.update(x, w)
        C.update(y, w)

    D = C - B
    assert math.isclose(D.get(), A.get())

    C -= B
    assert math.isclose(C.get(), A.get())


@pytest.mark.parametrize(
    "stat",
    [
        pytest.param(stat, id=stat.__class__.__name__)
        for stat in [stats.Mean(), stats.Var(ddof=0), stats.Var(ddof=1), stats.Var(ddof=2)]
    ],
)
def test_sub_back_to_zero(stat):
    A = copy.deepcopy(stat)
    B = copy.deepcopy(stat)
    C = copy.deepcopy(stat)

    x = random.random()
    A.update(x)
    B.update(x)

    D = A - B
    assert math.isclose(D.get(), C.get())


@pytest.mark.parametrize(
    "ddof",
    [pytest.param(ddof, id=f"{ddof=}") for ddof in [0, 1]],
)
def test_add_cov(ddof):
    c1 = stats.Cov(ddof=ddof)
    c2 = stats.Cov(ddof=ddof)

    X = [random.random() for _ in range(30)]
    Y = [random.random() for _ in range(30)]

    W = [random.random() for _ in range(30)]
    Z = [random.random() for _ in range(30)]

    for i, (x, y, w, z) in enumerate(zip(X, Y, W, Z)):
        c1.update(x, y)
        c2.update(w, z)
        if i >= c1.ddof:
            assert math.isclose(
                c1.get(),
                np.cov(X[: i + 1], Y[: i + 1], ddof=ddof)[0, 1],
                abs_tol=1e-10,
            )
            assert math.isclose(
                c2.get(),
                np.cov(W[: i + 1], Z[: i + 1], ddof=ddof)[0, 1],
                abs_tol=1e-10,
            )
            assert math.isclose(
                (c1 + c2).get(),
                np.cov(X[: i + 1] + W[: i + 1], Y[: i + 1] + Z[: i + 1], ddof=ddof)[0, 1],
                abs_tol=1e-10,
            )

    c1 += c2
    assert math.isclose(c1.get(), np.cov(X + W, Y + Z, ddof=ddof)[0, 1], abs_tol=1e-10)


@pytest.mark.parametrize(
    "ddof",
    [pytest.param(ddof, id=f"{ddof=}") for ddof in [0, 1]],
)
def test_sub_cov(ddof):
    c1 = stats.Cov(ddof=ddof)
    c2 = stats.Cov(ddof=ddof)

    X = [random.random() for _ in range(30)]
    Y = [random.random() for _ in range(30)]

    for x, y in zip(X, Y):
        c1.update(x, y)

    for i, (x, y) in enumerate(zip(X, Y)):
        c2.update(x, y)
        if i < len(X) - 2:
            assert math.isclose(
                (c1 - c2).get(), np.cov(X[i + 1 :], Y[i + 1 :], ddof=ddof)[0, 1], abs_tol=1e-10
            )

    # Test inplace subtraction
    X.extend(random.random() for _ in range(3))
    Y.extend(random.random() for _ in range(3))
    for i in range(30, 33):
        c1.update(X[i], Y[i])

    c1 -= c2
    assert math.isclose(c1.get(), np.cov(X[30:33], Y[30:33], ddof=ddof)[0, 1], abs_tol=1e-10)



from __future__ import annotations

import math
import random

from river import stats


def test_weighted_variance_with_close_numbers():
    """

    Origin of this test: https://github.com/online-ml/river/issues/732

    This test would fail if Var were implemented with a numerically unstable algorithm.

    """

    D = [
        (99.99999978143265, 6),
        (99.99999989071631, 8),
        (99.99999994535816, 6),
        (99.99999997267908, 9),
        (99.99999998633952, 10),
        (99.99999999316977, 3),
        (99.99999999829245, 5),
        (99.99999999957309, 9),
    ]

    var = stats.Var()

    for x, w in D:
        var.update(x, w)

    assert var.get() > 0 and math.isclose(var.get(), 4.648047194845607e-15)


def test_revert():
    for _ in range(5):
        X = [random.random() for _ in range(20)]

        v1 = stats.Var()
        v2 = stats.Var()

        for x in X[:10]:
            v1.update(x)
            v2.update(x)

        for x in X[10:]:
            v2.update(x)
        for x in X[10:]:
            v2.revert(x)

        assert math.isclose(v1.get(), v2.get())



from __future__ import annotations

import gymnasium as gym
import gymnasium.utils.env_checker
import pytest

from river import bandit


def _iter_envs():
    for env_name in gym.envs.registry:
        if env_name.startswith(bandit.envs.RIVER_NAMESPACE):
            yield gym.make(env_name)


@pytest.mark.parametrize(
    "env",
    [pytest.param(env, id=env.unwrapped.__class__.__name__) for env in _iter_envs()],
)
def test_gym_check_env(env):
    gym.utils.env_checker.check_env(env.unwrapped)



from __future__ import annotations

import math
import random

import numpy as np
import pandas as pd
import pytest

from river import covariance, stream


@pytest.mark.parametrize(
    "ddof",
    [
        pytest.param(
            ddof,
            id=f"{ddof=}",
        )
        for ddof in (0, 1, 2)
    ],
)
def test_covariance_revert(ddof):
    X = np.random.random((100, 5))
    X1 = X[: len(X) // 2]
    X2 = X[len(X) // 2 :]

    C1 = covariance.EmpiricalCovariance(ddof=ddof)
    for x, _ in stream.iter_array(X1):
        C1.update(x)

    C2 = covariance.EmpiricalCovariance(ddof=ddof)
    for x, _ in stream.iter_array(X):
        C2.update(x)
    for x, _ in stream.iter_array(X2):
        C2.revert(x)

    for k in C1._cov:
        assert math.isclose(C1._cov[k].get(), C2._cov[k].get())


@pytest.mark.parametrize(
    "ddof",
    [
        pytest.param(
            ddof,
            id=f"{ddof=}",
        )
        for ddof in (0, 1, 2)
    ],
)
def test_covariance_update_shuffled(ddof):
    C1 = covariance.EmpiricalCovariance(ddof=ddof)
    C2 = covariance.EmpiricalCovariance(ddof=ddof)

    X = np.random.random((100, 5))

    for x, _ in stream.iter_array(X):
        C1.update(x)
        C2.update({i: x[i] for i in random.sample(list(x.keys()), k=len(x))})

    for i, j in C1._cov:
        assert math.isclose(C1[i, j].get(), C2[i, j].get())


def test_covariance_update_sampled():
    # NOTE: this test only works with ddof=1 because pandas ignores it if there are missing values
    ddof = 1
    cov = covariance.EmpiricalCovariance(ddof=ddof)

    X = np.random.random((100, 5))
    samples = []

    for x, _ in stream.iter_array(X):
        sample = {i: x[i] for i in random.sample(list(x.keys()), k=len(x) - 1)}
        cov.update(sample)
        samples.append(sample)

    pd_cov = pd.DataFrame(samples).cov(ddof=ddof)

    for i, j in cov._cov:
        assert math.isclose(cov[i, j].get(), pd_cov.loc[i, j])


@pytest.mark.parametrize(
    "ddof",
    [
        pytest.param(
            ddof,
            id=f"{ddof=}",
        )
        for ddof in [0, 1]
    ],
)
def test_covariance_update_many(ddof):
    cov = covariance.EmpiricalCovariance(ddof=ddof)
    p = 5
    X_all = None

    for _ in range(p):
        n = np.random.randint(1, 31)
        X = pd.DataFrame(np.random.random((n, p)))

        cov.update_many(X)

        X_all = pd.concat((X_all, X)).astype(float)
        pd_cov = X_all.cov(ddof=ddof)

        for i, j in cov._cov:
            assert math.isclose(cov[i, j].get(), pd_cov.loc[i, j])


@pytest.mark.parametrize(
    "ddof",
    [
        pytest.param(
            ddof,
            id=f"{ddof=}",
        )
        for ddof in [0, 1]
    ],
)
def test_covariance_update_many_shuffled(ddof):
    cov = covariance.EmpiricalCovariance(ddof=ddof)
    p = 5
    X_all = None

    for _ in range(p):
        n = np.random.randint(5, 31)
        X = pd.DataFrame(np.random.random((n, p))).sample(p, axis="columns")

        cov.update_many(X)

        X_all = pd.concat((X_all, X)).astype(float)
        pd_cov = X_all.cov(ddof=ddof)

        for i, j in cov._cov:
            assert math.isclose(cov[i, j].get(), pd_cov.loc[i, j])


def test_covariance_update_many_sampled():
    # NOTE: this test only works with ddof=1 because pandas ignores it if there are missing values
    ddof = 1
    cov = covariance.EmpiricalCovariance(ddof=ddof)
    p = 5
    X_all = None

    for _ in range(p):
        n = np.random.randint(5, 31)
        X = pd.DataFrame(np.random.random((n, p))).sample(p - 1, axis="columns")

        cov.update_many(X)

        X_all = pd.concat((X_all, X)).astype(float)
        pd_cov = X_all.cov(ddof=ddof)

        for i, j in cov._cov:
            assert math.isclose(cov[i, j].get(), pd_cov.loc[i, j])


def test_precision_update_shuffled():
    C1 = covariance.EmpiricalPrecision()
    C2 = covariance.EmpiricalPrecision()

    X = np.random.random((100, 5))

    for x, _ in stream.iter_array(X):
        C1.update(x)
        C2.update({i: x[i] for i in random.sample(list(x.keys()), k=len(x))})

    for i, j in C1._inv_cov:
        assert math.isclose(C1[i, j], C2[i, j])


def test_precision_update_many_mini_batches():
    C1 = covariance.EmpiricalPrecision()
    C2 = covariance.EmpiricalPrecision()

    X = pd.DataFrame(np.random.random((100, 5)))

    C1.update_many(X)
    for Xb in np.split(X, 5):
        C2.update_many(Xb)

    for i, j in C1._inv_cov:
        assert math.isclose(C1[i, j], C2[i, j])


def test_precision_one_many_same():
    one = covariance.EmpiricalPrecision()
    many = covariance.EmpiricalPrecision()

    X = np.random.random((100, 5))

    for x, _ in stream.iter_array(X):
        one.update(x)
    many.update_many(pd.DataFrame(X))

    for i, j in one._inv_cov:
        assert math.isclose(one[i, j], many[i, j])



from __future__ import annotations

from river import datasets, metrics, stats, time_series


class MeanForecaster(time_series.base.Forecaster):
    def __init__(self):
        self.mean = stats.Mean()

    def learn_one(self, y, x=None):
        self.mean.update(y)

    def forecast(self, horizon, xs=None):
        return [self.mean.get()] * horizon


def test_forecasts_at_each_step():
    dataset = datasets.AirlinePassengers()
    model = MeanForecaster()
    metric = metrics.MAE()
    horizon = 12
    grace_period = 1

    steps = time_series.iter_evaluate(
        dataset=dataset, model=model, metric=metric, horizon=horizon, grace_period=grace_period
    )

    _, _, y_pred, _ = next(steps)
    assert y_pred == [112] * horizon
    _, _, y_pred, _ = next(steps)
    assert y_pred == [(112 + 118) / 2] * horizon
    _, _, y_pred, _ = next(steps)
    assert y_pred == [(112 + 118 + 132) / 3] * horizon
    _, _, y_pred, _ = next(steps)
    assert y_pred == [(112 + 118 + 132 + 129) / 4] * horizon

    n_steps = sum(
        1
        for _ in time_series.iter_evaluate(
            dataset=dataset, model=model, metric=metric, horizon=horizon, grace_period=grace_period
        )
    )
    assert n_steps == dataset.n_samples - horizon - grace_period



from __future__ import annotations

collect_ignore = []

try:
    import sklearn  # noqa: F401
except ImportError:
    collect_ignore.append("compat/test_sklearn.py")

try:
    import sqlalchemy  # noqa: F401
except ImportError:
    collect_ignore.append("stream/iter_sql.py")
    collect_ignore.append("stream/test_sql.py")

try:
    import vaex  # noqa: F401
except ImportError:
    collect_ignore.append("stream/iter_vaex.py")



from __future__ import annotations

from river import compose, datasets, linear_model, optim, preprocessing, stats, time_series


def test_clone_estimator():
    obj = linear_model.LinearRegression(l2=42)
    obj.learn_one({"x": 3}, 6)

    new = obj.clone({"l2": 21})
    assert new.l2 == 21
    assert obj.l2 == 42
    assert new.weights == {}
    assert new.weights != obj.weights


def test_clone_include_attributes():
    var = stats.Var()
    var.update(1)
    var.update(2)
    var.update(3)

    assert var._S == 2
    assert var.clone()._S == 0
    assert var.clone(include_attributes=True)._S == 2


def test_clone_pipeline():
    obj = preprocessing.StandardScaler() | linear_model.LinearRegression(l2=42)
    obj.learn_one({"x": 3}, 6)

    params = {"LinearRegression": {"l2": 21}}
    new = obj.clone(params)
    assert new["LinearRegression"].l2 == 21
    assert obj["LinearRegression"].l2 == 42
    assert new["LinearRegression"].weights == {}
    assert new["LinearRegression"].weights != obj["LinearRegression"].weights


def test_clone_idempotent():
    model = preprocessing.StandardScaler() | linear_model.LogisticRegression(
        optimizer=optim.Adam(), l2=0.1
    )

    trace = []
    for x, y in datasets.Phishing():
        trace.append(model.predict_proba_one(x))
        model.learn_one(x, y)

    clone = model.clone()
    for i, (x, y) in enumerate(datasets.Phishing()):
        assert clone.predict_proba_one(x) == trace[i]
        clone.learn_one(x, y)


def test_memory_usage():
    model = preprocessing.StandardScaler() | linear_model.LogisticRegression()

    # We can't test the exact value because it depends on the platform and the Python version
    # TODO: we could create a table of expected values for each platform and Python version
    assert isinstance(model._memory_usage, str)


def test_mutate():
    """

    >>> from river import datasets, linear_model, optim, preprocessing

    >>> model = (
    ...     preprocessing.StandardScaler() |
    ...     linear_model.LinearRegression(
    ...         optimizer=optim.SGD(3e-2)
    ...     )
    ... )

    >>> for x, y in datasets.TrumpApproval():
    ...     _ = model.predict_one(x)
    ...     model.learn_one(x, y)

    >>> len(model[-1].weights)
    6

    >>> model.mutate({'LinearRegression': {'optimizer': {'lr': {'learning_rate': 42}}}})
    >>> model[-1].optimizer
    SGD({'lr': Constant({'learning_rate': 42}), 'n_iterations': 1001})

    >>> model.mutate({'LinearRegression': {'optimizer': {'lr': optim.schedulers.Constant(43)}}})
    >>> model[-1].optimizer
    SGD({'lr': Constant({'learning_rate': 43}), 'n_iterations': 1001})

    >>> model.mutate({'LinearRegression': {'optimizer': optim.SGD(44)}})
    >>> model[-1].optimizer
    SGD({'lr': Constant({'learning_rate': 44}), 'n_iterations': 0})

    >>> model.mutate({'LinearRegression': {'l2': 0.123}})
    >>> model[-1].l2
    0.123

    >>> len(model[-1].weights)
    6

    >>> try:
    ...     model.mutate({'LinearRegression': {'weights': 'this is weird'}})
    ... except ValueError as e:
    ...     print(e)
    'weights' is not a mutable attribute of LinearRegression

    >>> try:
    ...     model.mutate({'LinearRegression': {'l3': 123}})
    ... except ValueError as e:
    ...     print(e)
    'l3' is not an attribute of LinearRegression

    """


def test_clone_positional_args():
    assert compose.Select(1, 2, 3).clone().keys == {1, 2, 3}
    assert compose.Discard("a", "b", "c").clone().keys == {"a", "b", "c"}
    assert compose.SelectType(float, int).clone().types == (float, int)


def test_clone_nested_pipeline():
    model = time_series.SNARIMAX(
        p=2,
        d=1,
        q=3,
        regressor=(
            preprocessing.StandardScaler()
            | linear_model.LinearRegression(optimizer=optim.SGD(3e-2))
        ),
    )
    assert model.clone()._get_params() == model._get_params()



from __future__ import annotations

import collections
import functools
import math
import random

import pandas as pd
import pytest
from sklearn import metrics as sk_metrics

from river import metrics


@pytest.mark.parametrize(
    "metric, sk_metric",
    [
        pytest.param(
            metric,
            sk_metric,
            id=f"{metric.__class__.__name__}"
            + (f"({metric.metric.__class__.__name__})" if hasattr(metric, "metric") else ""),
        )
        for metric, sk_metric in [
            (
                metrics.multioutput.ExactMatch(),
                sk_metrics.accuracy_score,
            ),
            (
                metrics.multioutput.MacroAverage(metrics.Precision()),
                functools.partial(sk_metrics.precision_score, average="macro", zero_division=0),
            ),
            (
                metrics.multioutput.MicroAverage(metrics.Precision()),
                functools.partial(sk_metrics.precision_score, average="micro", zero_division=0),
            ),
            (
                metrics.multioutput.SampleAverage(metrics.Precision()),
                functools.partial(sk_metrics.precision_score, average="samples", zero_division=0),
            ),
            (
                metrics.multioutput.MacroAverage(metrics.Recall()),
                functools.partial(sk_metrics.recall_score, average="macro", zero_division=0),
            ),
            (
                metrics.multioutput.MicroAverage(metrics.Recall()),
                functools.partial(sk_metrics.recall_score, average="micro", zero_division=0),
            ),
            (
                metrics.multioutput.SampleAverage(metrics.Recall()),
                functools.partial(sk_metrics.recall_score, average="samples", zero_division=0),
            ),
            (
                metrics.multioutput.MacroAverage(metrics.F1()),
                functools.partial(sk_metrics.f1_score, average="macro", zero_division=0),
            ),
            (
                metrics.multioutput.MicroAverage(metrics.F1()),
                functools.partial(sk_metrics.f1_score, average="micro", zero_division=0),
            ),
            (
                metrics.multioutput.SampleAverage(metrics.F1()),
                functools.partial(sk_metrics.f1_score, average="samples", zero_division=0),
            ),
        ]
    ],
)
def test_multiout_binary_clf(metric, sk_metric):
    y_true = []
    y_pred = []
    for _ in range(30):
        y_true.append({i: random.random() < 0.3 for i in range(3)})
        y_pred.append({i: random.random() < 0.3 for i in range(3)})
    y_true = pd.DataFrame(y_true)
    y_pred = pd.DataFrame(y_pred)

    for i, (yt, yp) in enumerate(
        zip(y_true.to_dict(orient="records"), y_pred.to_dict(orient="records"))
    ):
        metric.update(yt, yp)
        if i == 0:
            continue

        A = metric.get()
        B = sk_metric(y_true[: i + 1], y_pred[: i + 1])
        assert math.isclose(A, B, abs_tol=1e-3)


@pytest.mark.parametrize(
    "metric, sk_metric",
    [
        pytest.param(
            metric,
            sk_metric,
            id=f"{metric.__class__.__name__}"
            + (f"({metric.metric.__class__.__name__})" if hasattr(metric, "metric") else ""),
        )
        for metric, sk_metric in [
            # Both of the following cases are the same, because the average of averages is the same
            # as the global average
            (
                metrics.multioutput.MacroAverage(metrics.MAE()),
                functools.partial(sk_metrics.mean_absolute_error, multioutput="uniform_average"),
            ),
            (
                metrics.multioutput.MicroAverage(metrics.MAE()),
                functools.partial(sk_metrics.mean_absolute_error, multioutput="uniform_average"),
            ),
            #
            (
                metrics.multioutput.PerOutput(metrics.MAE()),
                functools.partial(sk_metrics.mean_absolute_error, multioutput="raw_values"),
            ),
        ]
    ],
)
def test_multiout_regression(metric, sk_metric):
    y_true = []
    y_pred = []
    for _ in range(30):
        y_true.append({i: random.random() for i in range(3)})
        y_pred.append({i: random.random() for i in range(3)})
    y_true = pd.DataFrame(y_true)
    y_pred = pd.DataFrame(y_pred)

    for i, (yt, yp) in enumerate(
        zip(y_true.to_dict(orient="records"), y_pred.to_dict(orient="records"))
    ):
        metric.update(yt, yp)
        if i == 0:
            continue

        A = metric.get()
        B = sk_metric(y_true[: i + 1], y_pred[: i + 1])

        if isinstance(A, collections.abc.Mapping):
            for k in A:
                assert math.isclose(A[k].get(), B[k], abs_tol=1e-3)
        else:
            assert math.isclose(A, B, abs_tol=1e-3)



from __future__ import annotations

import numpy as np
import pandas as pd
from sklearn import neighbors

from river import anomaly, datasets

np.random.seed(42)


def test_incremental_lof_scores():
    """
    Test that the incremental LOF algorithm returns similar LOF scores for each observation
    compared with the original static LOF algorithm implemented in scikit-learn.
    """
    norm_dist = 0.5 * np.random.rand(100, 2)
    x_inliers = np.concatenate((norm_dist - 2, norm_dist, norm_dist + 2), axis=0)
    x_outliers = np.concatenate(
        (
            np.random.uniform(low=-4, high=4, size=(20, 2)),
            np.random.uniform(low=-10, high=-5, size=(10, 2)),
            np.random.uniform(low=5, high=10, size=(10, 2)),
        ),
        axis=0,
    )
    x_train = np.concatenate((x_inliers, x_outliers), axis=0)
    x_train_dict = [{f"feature_{i + 1}": elem[i] for i in range(2)} for elem in x_train]
    ground_truth = np.ones(len(x_train), dtype=int)
    ground_truth[-len(x_outliers) :] = -1
    df_train = pd.DataFrame({"observations": x_train_dict, "ground_truth": ground_truth})
    x_pred = np.random.uniform(low=-5, high=5, size=(30, 2))
    x_pred_dict = [{f"feature_{i + 1}": elem[i] for i in range(2)} for elem in x_pred]
    incremental_lof = anomaly.LocalOutlierFactor(n_neighbors=20)

    for x in df_train["observations"]:
        incremental_lof.learn_one(x)

    ilof_scores_train = np.array([ilof_score for ilof_score in incremental_lof.lof.values()])

    ilof_scores_pred = []
    for x in x_pred_dict:
        ilof_scores_pred.append(incremental_lof.score_one(x))

    lof_sklearn = neighbors.LocalOutlierFactor(n_neighbors=20)
    lof_sklearn.fit_predict(x_train)
    lof_sklearn_scores_train = -lof_sklearn.negative_outlier_factor_

    assert np.allclose(ilof_scores_train, lof_sklearn_scores_train, rtol=1e-08, atol=1e-08)


def test_batch_lof_scores():
    """
    Test that the incremental LOF algorithm returns similar LOF scores for each batch
    with `learn_many` compared with the original static LOF algorithm implemented in scikit-learn,
    under different batch sizes.
    """
    cc_df = pd.DataFrame(datasets.CreditCard())
    cc_df_np = [np.array(list(x.values())) for x in cc_df[0].to_dict().values()]

    batch_sizes = [20, 50, 100]

    for batch_size in batch_sizes:
        ilof_river_batch = anomaly.LocalOutlierFactor(n_neighbors=20)
        ilof_river_batch.learn_many(cc_df[0:batch_size])
        ilof_scores_river_batch = np.array([v for v in ilof_river_batch.lof.values()])

        lof_sklearn_batch = neighbors.LocalOutlierFactor(n_neighbors=20)
        lof_sklearn_batch.fit_predict(cc_df_np[0:batch_size])
        lof_scores_sklearn_batch = -lof_sklearn_batch.negative_outlier_factor_

        assert np.allclose(
            ilof_scores_river_batch, lof_scores_sklearn_batch, rtol=1e-02, atol=1e-02
        )


def test_issue_1328():
    lof = anomaly.LocalOutlierFactor()
    X = [{"a": 1, "b": 1}, {"a": 1, "b": 1}]
    for x in X:
        lof.learn_one(x)


def test_issue_1331():
    import copy

    from river import anomaly

    lof = anomaly.LocalOutlierFactor()

    X = [{"a": 1, "b": 1}, {"a": 1, "b": 1}]
    for x in X:
        lof.learn_one(x)

    neighborhoods_ = lof.neighborhoods.copy()
    rev_neighborhoods = lof.rev_neighborhoods.copy()
    k_dist_ = lof.k_dist.copy()
    reach_dist_ = copy.deepcopy(lof.reach_dist)
    dist_dict_ = copy.deepcopy(lof.dist_dict)
    local_reach_ = lof.local_reach.copy()
    lof_ = lof.lof.copy()

    lof.score_one({"a": 0.5, "b": 1})

    assert neighborhoods_ == lof.neighborhoods
    assert rev_neighborhoods == lof.rev_neighborhoods
    assert k_dist_ == lof.k_dist
    assert reach_dist_ == lof.reach_dist
    assert dist_dict_ == lof.dist_dict
    assert local_reach_ == lof.local_reach
    assert lof_ == lof.lof



from __future__ import annotations

import numpy as np
import pandas as pd
from sklearn import datasets as sk_datasets
from sklearn import linear_model as sk_linear_model

from river import compat, compose, datasets, preprocessing, stream


def test_issue_1238():
    """

    https://github.com/online-ml/river/issues/1238

    >>> from river import compose

    >>> x = dict(a=0, b=1, x=2, y=3)
    >>> group_1 = compose.Select('a', 'b')
    >>> group_2 = compose.Select('x', 'y')
    >>> product = group_1 + group_2 + group_1 * group_2
    >>> sorted(product.transform_one(x), key=lambda k: (len(k), k))
    ['a', 'b', 'x', 'y', 'a*x', 'a*y', 'b*x', 'b*y']

    >>> product = group_1 * group_2 + group_1 + group_2
    >>> sorted(product.transform_one(x), key=lambda k: (len(k), k))
    ['a', 'b', 'x', 'y', 'a*x', 'a*y', 'b*x', 'b*y']

    >>> product = group_1 + group_1 * group_2
    >>> sorted(product.transform_one(x), key=lambda k: (len(k), k))
    ['a', 'b', 'a*x', 'a*y', 'b*x', 'b*y']

    """


def test_issue_1243():
    """

    https://github.com/online-ml/river/issues/1243

    >>> import random
    >>> import pandas as pd
    >>> from river import compose, preprocessing

    >>> rng = random.Random(42)
    >>> X = [{'x': rng.uniform(8, 12), 'z': 1 } for _ in range(6)]

    >>> X = pd.DataFrame.from_dict(X)
    >>> group1 = compose.Select('z')
    >>> group2 = compose.Select('x') | preprocessing.StandardScaler()
    >>> model = group1 + group2 + group1 * group2
    >>> model.learn_many(X)
    >>> for x in X.to_dict('records'):
    ...     print(model.transform_one(x))
    {'z*x': 0.785..., 'x': 0.785..., 'z': 1}
    {'z*x': -1.511..., 'x': -1.511..., 'z': 1}
    {'z*x': -0.576..., 'x': -0.576..., 'z': 1}
    {'z*x': -0.770..., 'x': -0.770..., 'z': 1}
    {'z*x': 1.148..., 'x': 1.148..., 'z': 1}
    {'z*x': 0.924..., 'x': 0.924..., 'z': 1}

    """


def test_issue_1253():
    """

    https://github.com/online-ml/river/issues/1253

    >>> import numpy as np
    >>> import pandas as pd
    >>> from river import compat, compose, preprocessing
    >>> from sklearn import datasets, linear_model

    >>> np.random.seed(1000)
    >>> X, y = datasets.make_regression(n_samples=5_000, n_features=2)
    >>> X = pd.DataFrame(X, columns=['feat_1','feat_2'])
    >>> X['cat'] = np.random.randint(1, 100, len(X))
    >>> X['cat'] = X['cat'].astype('string')

    >>> group1 = compose.Select('cat') | preprocessing.OneHotEncoder()
    >>> group2 = compose.Select('feat_2') | preprocessing.StandardScaler()
    >>> model = group1 + group1 * group2
    >>> XT = model.transform_many(X)

    >>> XT.memory_usage().sum().item() // 1000
    85

    >>> XT.sparse.to_dense().memory_usage().sum().item() // 1000
    4455

    >>> X, y = datasets.make_regression(n_samples=6, n_features=2)
    >>> X = pd.DataFrame(X)
    >>> X.columns = ['feat_1','feat_2']
    >>> X['cat'] = np.random.randint(1, 3, X.shape[0])
    >>> y = pd.Series(y)
    >>> group1 = compose.Select('cat') | preprocessing.OneHotEncoder()
    >>> group2 = compose.Select('feat_2') | preprocessing.StandardScaler()
    >>> sparsify = lambda X: X.astype({
    ...     key: pd.SparseDtype(X.dtypes[key].type, fill_value=0)
    ...     for key in X.dtypes.keys()
    ... })
    >>> model = (
    ...     (group1 + group1 * group2) |
    ...     compose.FuncTransformer(sparsify) |
    ...     compat.convert_sklearn_to_river(linear_model.SGDRegressor(max_iter=3))
    ... )
    >>> _ = model.predict_many(X)
    >>> model.transform_many(X)
       cat_1*feat_2  cat_2*feat_2  cat_1  cat_2
    0     -1.196841             0      1      0
    1      1.304619             0      1      0
    2     -1.294091             0      1      0
    3      0.287426             0      1      0
    4     -0.143960             0      1      0
    5             0      1.042847      0      1

    """


def test_left_is_pipeline():
    group_1 = compose.Select("a", "b")
    group_2 = compose.Select("x", "y") | preprocessing.OneHotEncoder()

    product = group_1 + group_2 + group_1 * group_2
    assert product.transform_one(dict(a=1, b=2, x=4, y=4, z=5)) == {
        "a*y_4": 1,
        "a*x_4": 1,
        "b*y_4": 2,
        "b*x_4": 2,
        "y_4": 1,
        "x_4": 1,
        "a": 1,
        "b": 2,
    }


def test_right_is_pipeline():
    group_1 = compose.Select("a", "b") | preprocessing.OneHotEncoder()
    group_2 = compose.Select("x", "y")

    product = group_1 + group_2 + group_1 * group_2
    assert product.transform_one(dict(a=1, b=2, x=4, y=4, z=5)) == {
        "a_1*x": 4,
        "a_1*y": 4,
        "b_2*x": 4,
        "b_2*y": 4,
        "x": 4,
        "y": 4,
        "a_1": 1,
        "b_2": 1,
    }


def test_both_are_pipelines():
    group_1 = compose.Select("a", "b") | preprocessing.OneHotEncoder()
    group_2 = compose.Select("x", "y") | preprocessing.OneHotEncoder()

    product = group_1 + group_2 + group_1 * group_2
    assert product.transform_one(dict(a=1, b=2, x=4, y=4, z=5)) == {
        "b_2*x_4": 1,
        "b_2*y_4": 1,
        "a_1*x_4": 1,
        "a_1*y_4": 1,
        "x_4": 1,
        "y_4": 1,
        "b_2": 1,
        "a_1": 1,
    }


def test_renaming():
    renamer = compose.Renamer(dict(a="z", b="y", c="x"))
    assert renamer.transform_one(dict(a=1, b=2, d=3)) == dict(z=1, y=2, d=3)


def test_prefixing():
    prefixer = compose.Prefixer("x_")
    assert prefixer.transform_one(dict(a=1, b=2, d=3)) == dict(x_a=1, x_b=2, x_d=3)


def test_suffixing():
    suffixer = compose.Suffixer("_x")
    assert suffixer.transform_one(dict(a=1, b=2, d=3)) == dict(a_x=1, b_x=2, d_x=3)


def test_one_many_consistent():
    """Checks that using transform_one or transform_many produces the same result."""

    product = (
        compose.Select("ordinal_date")
        + compose.Select("gallup", "ipsos") * compose.Select("morning_consult")
        + compose.Select("rasmussen") * compose.Select("you_gov", "five_thirty_eight")
    )
    X = pd.read_csv(datasets.TrumpApproval().path)

    one_outputs = []
    for x, _ in stream.iter_pandas(X):
        one_outputs.append(product.transform_one(x))
    one_outputs = pd.DataFrame(one_outputs)

    many_outputs = product.transform_many(X)

    # check_dtype=False to avoid int/float comparison
    pd.testing.assert_frame_equal(many_outputs[one_outputs.columns], one_outputs, check_dtype=False)


def test_issue_1310():
    X, y = sk_datasets.make_regression(n_samples=5000, n_features=2)
    X = pd.DataFrame(X)
    X.columns = ["feat_1", "feat_2"]
    X["cat"] = np.random.randint(1, 100, X.shape[0])
    X["cat"] = X["cat"].astype("string")
    y = pd.Series(y)

    group1 = compose.Select("cat") | preprocessing.OneHotEncoder()
    group2 = compose.Select("feat_2") | preprocessing.StandardScaler()
    model = group1 + group1 * group2 * group2 | compat.convert_sklearn_to_river(
        sk_linear_model.SGDRegressor()
    )

    model.predict_many(X)
    model.learn_many(X, y)



from __future__ import annotations

import pandas as pd
import pytest
from sklearn import datasets as sk_datasets
from sklearn import linear_model as sk_linear_model
from sklearn.utils import estimator_checks

from river import base, cluster, compat, linear_model, preprocessing


@pytest.mark.parametrize(
    "estimator",  # type: ignore[misc]
    [
        pytest.param(estimator, id=str(estimator))
        for estimator in [
            linear_model.LinearRegression(),
            linear_model.LogisticRegression(),
            preprocessing.StandardScaler(),
            cluster.KMeans(n_clusters=3, seed=42),
        ]
    ],
)
@pytest.mark.filterwarnings("ignore::sklearn.utils.estimator_checks.SkipTestWarning")  # type: ignore[misc]
def test_river_to_sklearn_check_estimator(estimator: base.Estimator):
    skl_estimator = compat.convert_river_to_sklearn(estimator)
    estimator_checks.check_estimator(skl_estimator)


@pytest.mark.filterwarnings("ignore::sklearn.utils.estimator_checks.SkipTestWarning")
def test_sklearn_check_twoway():
    estimator = sk_linear_model.SGDRegressor()
    river_estimator = compat.convert_sklearn_to_river(estimator)
    skl_estimator = compat.convert_river_to_sklearn(river_estimator)
    estimator_checks.check_estimator(skl_estimator)


@pytest.mark.parametrize(
    "estimator",
    [
        pytest.param(estimator, id=str(estimator))
        for estimator in [
            compat.convert_sklearn_to_river(sk_linear_model.SGDRegressor()),
            (
                preprocessing.StandardScaler()
                | compat.convert_sklearn_to_river(sk_linear_model.SGDRegressor())
            ),
        ]
    ],
)
def test_not_fitted_still_works_regression(estimator):
    X, _ = sk_datasets.make_regression(n_samples=500, n_features=4)
    X = pd.DataFrame(X)
    y_pred = estimator.predict_many(X)
    assert len(y_pred) == len(X)
    assert y_pred.eq(0).all()


@pytest.mark.parametrize(
    "estimator,n_classes",
    [
        pytest.param(estimator, n_classes, id=f"{estimator}-{n_classes} classes")
        for n_classes in [2, 3]
        for estimator in [
            compat.convert_sklearn_to_river(
                sk_linear_model.SGDClassifier(loss="log_loss"), classes=list(range(n_classes))
            ),
            (
                preprocessing.StandardScaler()
                | compat.convert_sklearn_to_river(
                    sk_linear_model.SGDClassifier(loss="log_loss"), classes=list(range(n_classes))
                )
            ),
        ]
    ],
)
def test_not_fitted_still_works_classification(estimator, n_classes):
    X, _ = sk_datasets.make_classification(
        n_samples=500, n_features=10, n_informative=6, n_classes=n_classes
    )
    X = pd.DataFrame(X)
    y_pred = estimator.predict_many(X)
    assert len(y_pred) == len(X)
    assert y_pred.eq(0).all()

    y_pred_proba = estimator.predict_proba_many(X)
    assert y_pred_proba.shape == (len(X), n_classes)



from __future__ import annotations

import copy
import importlib
import inspect
import random

import gymnasium as gym
import pytest

from river import bandit, metrics

from .test_envs import _iter_envs


def test_ranking():
    class DummyMetric(metrics.base.Metric):
        def __init__(self):
            self.value = None

        def get(self):
            return self.value

        @property
        def bigger_is_better(self):
            return False

        def revert(self): ...

        def update(self): ...

        def works_with(self): ...

    class DummyPolicy(bandit.base.Policy):
        def __init__(self):
            super().__init__(reward_obj=DummyMetric())

        def _pull(self, arms): ...

    policy = DummyPolicy()
    policy._rewards[0].value = 0
    policy._rewards[1].value = 1
    policy._rewards[2].value = 2
    assert policy.ranking == [0, 1, 2]

    policy._rewards[0].value = 2
    policy._rewards[1].value = 1
    policy._rewards[2].value = 0
    assert policy.ranking == [2, 1, 0]

    policy._rewards[0].value = 0
    policy._rewards[1].value = 2
    policy._rewards[2].value = 1
    assert policy.ranking == [0, 2, 1]


class RandomPolicy(bandit.base.Policy):
    def _pull(self, arm_ids):
        return random.choice(arm_ids)


def _iter_policies():
    for _, policy in inspect.getmembers(importlib.import_module("river.bandit"), inspect.isclass):
        for params in policy._unit_test_params():
            yield policy(**params)


@pytest.mark.parametrize(  # type: ignore[misc]
    "policy,env",
    [
        pytest.param(
            policy, env, id=f"{policy.__class__.__name__}-{env.unwrapped.__class__.__name__}"
        )
        # TODO: add simpler environments to test with
        for policy in _iter_policies()
        for env in _iter_envs()
    ],
)
@pytest.mark.skip(reason="flaky")  # type: ignore[misc]
def test_better_than_random_policy(policy: bandit.base.Policy, env: gym.Env):
    """Test that the policy is better than random."""

    n_trials = 30
    n_successes = 0

    for _ in range(n_trials):
        policy = policy.clone()
        random_policy = RandomPolicy()

        env_seed = random.randint(0, 2**32 - 1)
        env = copy.deepcopy(env)
        _ = env.reset(seed=env_seed)
        _ = env.action_space.seed(env_seed)

        random_env = copy.deepcopy(env)
        _ = random_env.reset(seed=env_seed)
        _ = random_env.action_space.seed(env_seed)

        policy_reward = 0.0
        random_reward = 0.0

        terminated, truncated = False, False

        arm_ids = list(range(env.action_space.n))  # type: ignore

        while not terminated and not truncated:
            arm_id = policy.pull(arm_ids)  # type: ignore
            observation, reward, terminated, truncated, info = env.step(arm_id)
            policy.update(arm_id, reward)
            policy_reward += float(reward)

            random_arm_id = random_policy.pull(arm_ids)  # type: ignore
            (
                observation,
                reward,
                terminated,
                truncated,
                info,
            ) = random_env.step(random_arm_id)
            random_policy.update(random_arm_id, reward)
            random_reward += float(reward)

        n_successes += policy_reward > random_reward

    assert n_successes > n_trials * 0.8



from __future__ import annotations

import copy
import itertools
import math
import random

import numpy as np
import pandas as pd
import pytest
from sklearn import linear_model as sklm
from sklearn.datasets import make_classification, make_regression
from sklearn.metrics import log_loss

from river import datasets, optim, preprocessing, stream, utils
from river import linear_model as lm


def iter_perturbations(keys, n=10):
    """Enumerate perturbations that will be applied to the weights."""

    # Enumerate unit vectors
    for i in keys:
        yield {j: int(i == j) for j in keys}

    # Enumerate samples from a spherical Gaussian
    for _ in range(n):
        p = {j: random.gauss(0, 1) for j in keys}
        norm = utils.math.norm(p, order=2)
        for j in p:
            p[j] /= norm
        yield p


@pytest.mark.parametrize(
    "lm, dataset",
    [
        pytest.param(
            lm(optimizer=copy.deepcopy(optimizer), initializer=initializer, l2=0),
            dataset,
            id=f"{lm.__name__} - {optimizer} - {initializer}",
        )
        for lm, dataset in [
            (lm.LinearRegression, datasets.TrumpApproval().take(100)),
            (lm.LogisticRegression, datasets.Bananas().take(100)),
        ]
        for optimizer, initializer in itertools.product(
            [
                optim.AdaBound(),
                optim.AdaDelta(),
                optim.AdaGrad(),
                optim.AdaMax(),
                optim.Adam(),
                optim.AMSGrad(),
                # TODO: check momentum optimizers
                # optim.Momentum(),
                # optim.NesterovMomentum(),
                optim.RMSProp(),
                optim.SGD(),
            ],
            [
                optim.initializers.Zeros(),
                optim.initializers.Normal(mu=0, sigma=1, seed=42),
            ],
        )
    ],
)
def test_finite_differences(lm, dataset):
    """Checks the gradient of a linear model via finite differences.

    References
    ----------
    [^1]: [How to test gradient implementations](https://timvieira.github.io/blog/post/2017/04/21/how-to-test-gradient-implementations/)
    [^2]: [Stochastic Gradient Descent Tricks](https://cilvr.cs.nyu.edu/diglib/lsml/bottou-sgd-tricks-2012.pdf)

    """

    scaler = preprocessing.StandardScaler()
    eps = 1e-6

    for x, y in dataset:
        scaler.learn_one(x)
        x = scaler.transform_one(x)

        # Store the current gradient and weights
        gradient, _ = lm._eval_gradient_one(x, y, 1)
        weights = copy.deepcopy(lm._weights)

        # d is a set of weight perturbations
        for d in iter_perturbations(weights.keys()):
            # Pertubate the weights and obtain the loss with the new weights
            lm._weights = utils.VectorDict({i: weights[i] + eps * di for i, di in d.items()})
            forward = lm.loss(y_true=y, y_pred=lm._raw_dot_one(x))
            lm._weights = utils.VectorDict({i: weights[i] - eps * di for i, di in d.items()})
            backward = lm.loss(y_true=y, y_pred=lm._raw_dot_one(x))

            # We expect g and h to be equal
            g = utils.math.dot(d, gradient)
            h = (forward - backward) / (2 * eps)

            # Compare signs
            # TODO: reactivate this check
            # assert np.sign(g) == np.sign(h)

            # Check absolute difference
            # TODO: decrease the tolerance
            assert abs(g - h) < 1e-5

        # Reset the weights to their original values in order not to influence
        # the training loop, even though it doesn't really matter.
        lm._weights = weights
        lm.learn_one(x, y)


def test_one_many_consistent():
    """Checks that using learn_one or learn_many produces the same result."""

    X = pd.read_csv(datasets.TrumpApproval().path)
    Y = X.pop("five_thirty_eight")

    one = lm.LinearRegression()
    for x, y in stream.iter_pandas(X[:100], Y[:100]):
        one.learn_one(x, y)

    many = lm.LinearRegression()
    for xb, yb in zip(np.array_split(X[:100], len(X[:100])), np.array_split(Y[:100], len(Y[:100]))):
        many.learn_many(xb, yb)

    for i in X:
        assert math.isclose(one.weights[i], many.weights[i])


def test_shuffle_columns():
    """Checks that learn_many works identically whether columns are shuffled or not."""

    X = pd.read_csv(datasets.TrumpApproval().path)
    Y = X.pop("five_thirty_eight")

    normal = lm.LinearRegression()
    for xb, yb in zip(np.array_split(X, 10), np.array_split(Y, 10)):
        normal.learn_many(xb, yb)

    shuffled = lm.LinearRegression()
    for xb, yb in zip(np.array_split(X, 10), np.array_split(Y, 10)):
        cols = np.random.permutation(X.columns)
        shuffled.learn_many(xb[cols], yb)

    for i in X:
        assert math.isclose(normal.weights[i], shuffled.weights[i])


def test_add_remove_columns():
    """Checks that no exceptions are raised whenever columns are dropped and/or added."""

    X = pd.read_csv(datasets.TrumpApproval().path)
    Y = X.pop("five_thirty_eight")

    lin_reg = lm.LinearRegression()
    for xb, yb in zip(np.array_split(X, 10), np.array_split(Y, 10)):
        # Pick half of the columns at random
        cols = np.random.choice(X.columns, len(X.columns) // 2, replace=False)
        lin_reg.learn_many(xb[cols], yb)


class ScikitLearnSquaredLoss:
    """sklearn removes the leading 2 from the gradient of the squared loss."""

    def gradient(self, y_true, y_pred):
        return y_pred - y_true


lin_reg_tests = {
    "Vanilla": (
        {"optimizer": optim.SGD(1e-2), "loss": ScikitLearnSquaredLoss()},
        {"learning_rate": "constant", "eta0": 1e-2, "alpha": 0},
    ),
    "Huber": (
        {"optimizer": optim.SGD(1e-2), "loss": optim.losses.Huber()},
        {"loss": "huber", "learning_rate": "constant", "eta0": 1e-2, "alpha": 0},
    ),
    "No intercept": (
        {
            "optimizer": optim.SGD(1e-2),
            "intercept_lr": 0,
            "loss": ScikitLearnSquaredLoss(),
        },
        {
            "learning_rate": "constant",
            "eta0": 1e-2,
            "alpha": 0,
            "fit_intercept": False,
        },
    ),
    "L2 regu": (
        {
            "optimizer": optim.SGD(1e-2),
            "loss": ScikitLearnSquaredLoss(),
            "l2": 1e-3,
        },
        {"learning_rate": "constant", "eta0": 1e-2, "alpha": 1e-3},
    ),
}


@pytest.mark.parametrize(
    "river_params, sklearn_params",
    lin_reg_tests.values(),
    ids=lin_reg_tests.keys(),
)
def test_lin_reg_sklearn_coherence(river_params, sklearn_params):
    """Checks that the sklearn and river implementations produce the same results."""

    ss = preprocessing.StandardScaler()
    rv = lm.LinearRegression(**river_params)
    sk = sklm.SGDRegressor(**sklearn_params)

    for x, y in datasets.TrumpApproval().take(100):
        ss.learn_one(x)
        x = ss.transform_one(x)
        rv.learn_one(x, y)
        sk.partial_fit([list(x.values())], [y])

    for i, w in enumerate(rv.weights.values()):
        assert math.isclose(w, sk.coef_[i])

    assert math.isclose(rv.intercept, sk.intercept_[0])


@pytest.mark.parametrize(
    "river_params, sklearn_params",
    lin_reg_tests.values(),
    ids=lin_reg_tests.keys(),
)
def test_lin_reg_sklearn_learn_many_coherence(river_params, sklearn_params):
    """Checks that the sklearn and river implementations produce the same results
    when learn_many is used."""

    ss = preprocessing.StandardScaler()
    rv = lm.LinearRegression(**river_params)
    sk = sklm.SGDRegressor(**sklearn_params)

    for x, y in datasets.TrumpApproval().take(100):
        ss.learn_one(x)
        x = ss.transform_one(x)
        rv.learn_many(pd.DataFrame([x]), pd.Series([y]))
        sk.partial_fit([list(x.values())], [y])

    for i, w in enumerate(rv.weights.values()):
        assert math.isclose(w, sk.coef_[i])

    assert math.isclose(rv.intercept, sk.intercept_[0])


log_reg_tests = {
    "Vanilla": (
        {"optimizer": optim.SGD(1e-2)},
        {"learning_rate": "constant", "eta0": 1e-2, "alpha": 0, "loss": "log_loss"},
    ),
    "Hinge": (
        {"optimizer": optim.SGD(1e-2), "loss": optim.losses.Hinge()},
        {"learning_rate": "constant", "eta0": 1e-2, "alpha": 0},
    ),
    "No intercept": (
        {"optimizer": optim.SGD(1e-2), "intercept_lr": 0},
        {
            "learning_rate": "constant",
            "eta0": 1e-2,
            "alpha": 0,
            "loss": "log_loss",
            "fit_intercept": False,
        },
    ),
    "L2 regu": (
        {
            "optimizer": optim.SGD(1e-2),
            "l2": 1e-3,
        },
        {
            "learning_rate": "constant",
            "eta0": 1e-2,
            "alpha": 1e-3,
            "loss": "log_loss",
        },
    ),
    "Inverse-scaling": (
        {
            "optimizer": optim.SGD(optim.schedulers.InverseScaling(1e-2)),
            "intercept_lr": optim.schedulers.InverseScaling(1e-2),
        },
        {"eta0": 1e-2, "alpha": 0, "learning_rate": "invscaling", "loss": "log_loss"},
    ),
    "Optimal": (
        {
            "optimizer": optim.SGD(optim.schedulers.Optimal(optim.losses.Hinge(), alpha=1e-3)),
            "loss": optim.losses.Hinge(),
            "intercept_lr": optim.schedulers.Optimal(optim.losses.Hinge(), alpha=1e-3),
            "l2": 1e-3,
        },
        {"learning_rate": "optimal", "alpha": 1e-3},
    ),
    "Optimal no intercept": (
        {
            "optimizer": optim.SGD(optim.schedulers.Optimal(optim.losses.Hinge(), alpha=1e-3)),
            "loss": optim.losses.Hinge(),
            "intercept_lr": 0,
            "l2": 1e-3,
        },
        {"learning_rate": "optimal", "alpha": 1e-3, "fit_intercept": False},
    ),
}


@pytest.mark.parametrize(
    "river_params, sklearn_params",
    log_reg_tests.values(),
    ids=log_reg_tests.keys(),
)
def test_log_reg_sklearn_coherence(river_params, sklearn_params):
    """Checks that the sklearn and river implementations produce the same results."""

    ss = preprocessing.StandardScaler()
    rv = lm.LogisticRegression(**river_params)
    sk = sklm.SGDClassifier(**sklearn_params)

    for x, y in datasets.Bananas().take(100):
        ss.learn_one(x)
        x = ss.transform_one(x)
        rv.learn_one(x, y)
        sk.partial_fit([list(x.values())], [y], classes=[False, True])

    for i, w in enumerate(rv.weights.values()):
        assert math.isclose(w, sk.coef_[0][i])

    assert math.isclose(rv.intercept, sk.intercept_[0])


perceptron_tests = {
    "Vanilla": (
        {},
        {},
    ),
    "L2 regu": (
        {
            "l2": 1e-3,
        },
        {
            "alpha": 1e-3,
            "penalty": "l2",
        },
    ),
}


@pytest.mark.parametrize(
    "river_params, sklearn_params",
    perceptron_tests.values(),
    ids=perceptron_tests.keys(),
)
def test_perceptron_sklearn_coherence(river_params, sklearn_params):
    """Checks that the sklearn and river implementations produce the same results."""

    ss = preprocessing.StandardScaler()
    rv = lm.Perceptron(**river_params)
    sk = sklm.Perceptron(**sklearn_params)

    for x, y in datasets.Bananas().take(100):
        ss.learn_one(x)
        x = ss.transform_one(x)
        rv.learn_one(x, y)
        sk.partial_fit([list(x.values())], [y], classes=[False, True])

    for i, w in enumerate(rv.weights.values()):
        assert math.isclose(w, sk.coef_[0][i])

    assert math.isclose(rv.intercept, sk.intercept_[0])


def test_lin_reg_sklearn_l1_non_regression():
    """Checks that the river L1 implementation results are no worse than sklearn L1."""

    X, y, true_coeffs = make_regression(
        n_samples=1000,
        n_features=20,
        n_informative=4,
        coef=True,
        random_state=273,
    )
    X = pd.DataFrame(X)
    y = pd.Series(y)

    ss = preprocessing.StandardScaler()

    rv = lm.LinearRegression(
        **{
            "optimizer": optim.SGD(1e-2),
            "loss": ScikitLearnSquaredLoss(),
            "l1": 1e-1,
        }
    )

    sk = sklm.SGDRegressor(
        **{
            "learning_rate": "constant",
            "eta0": 1e-2,
            "alpha": 1e-1,
            "penalty": "l1",
        }
    )

    for xi, yi in stream.iter_pandas(X, y):
        ss.learn_one(xi)
        xi_tr = ss.transform_one(xi)
        rv.learn_one(xi_tr, yi)
        sk.partial_fit([list(xi_tr.values())], [yi])

    rv_coeffs = np.array(list(rv.weights.values()))
    sk_coeffs = sk.coef_

    # check that the shrinkage to zero is no worse than in sklearn
    assert np.sum(rv_coeffs > 0) <= np.sum(sk_coeffs > 0)

    # check that the river coefficients shrink to the true ones with 10% relative tolerance
    assert np.isclose(rv_coeffs, true_coeffs, rtol=5e-02, atol=0.0).all()


def test_log_reg_sklearn_l1_non_regression():
    """Checks that the river L1 implementation results are no worse than sklearn L1."""

    (
        X,
        y,
    ) = make_classification(
        n_samples=1000,
        n_features=20,
        n_informative=4,
        n_classes=2,
        random_state=273,
    )
    X = pd.DataFrame(X)
    y = pd.Series(y)

    ss = preprocessing.StandardScaler()

    rv = lm.LogisticRegression(
        **{
            "optimizer": optim.SGD(1e-2),
            "l1": 1e-3,
        }
    )

    sk = sklm.SGDClassifier(
        **{
            "learning_rate": "constant",
            "eta0": 1e-2,
            "alpha": 1e-3,
            "penalty": "l1",
            "loss": "log_loss",
        }
    )

    rv_pred = list()
    sk_pred = list()
    for xi, yi in stream.iter_pandas(X, y):
        ss.learn_one(xi)
        xi_tr = ss.transform_one(xi)
        rv.learn_one(xi_tr, yi)
        sk.partial_fit([list(xi_tr.values())], [yi], classes=[False, True])

        rv_pred.append(rv.predict_one(xi_tr))
        sk_pred.append(sk.predict([list(xi_tr.values())])[0])

    rv_coeffs = np.array(list(rv.weights.values()))
    sk_coeffs = sk.coef_

    # check that the shrinkage to zero is no worse than in sklearn
    assert np.sum(rv_coeffs > 0) <= np.sum(sk_coeffs > 0)

    # since we can't access true coefficients from sklearn.make_classification, we compare losses
    assert math.isclose(log_loss(y, rv_pred), log_loss(y, sk_pred))



from __future__ import annotations

import math

from sklearn import metrics as sk_metrics

from river import metrics


def test_cross_entropy():
    metric = metrics.CrossEntropy()

    y_true = [0, 1, 2, 2]
    y_pred = [
        [0.29450637, 0.34216758, 0.36332605],
        [0.21290077, 0.32728332, 0.45981591],
        [0.42860913, 0.33380113, 0.23758974],
        [0.44941979, 0.32962558, 0.22095463],
    ]

    for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
        yp = dict(enumerate(yp))
        metric.update(yt, yp)

        if i >= 1:
            assert math.isclose(
                metric.get(),
                sk_metrics.log_loss(y_true[: i + 1], y_pred[: i + 1], labels=[0, 1, 2]),
            )

    metric.revert(y_true[-1], dict(enumerate(y_pred[-1])))
    assert math.isclose(metric.get(), sk_metrics.log_loss(y_true[:-1], y_pred[:-1]))



"""
The tests performed here confirm that the outputs of the online preprocessing.LDA are exactly the
same as those of the original with a batch_size of size 1. Coverage is 100%.

References
----------
[^1]: Jordan Boyd-Graber, Ke Zhai, Online Latent Dirichlet Allocation with Infinite Vocabulary.
    http://proceedings.mlr.press/v28/zhai13.pdf
[^2]: river's LDA implementation reproduces exactly the same results as the original one from
    PyInfVov (https://github.com/kzhai/PyInfVoc) with a batch size of 1.

"""

from __future__ import annotations

import numpy as np

from river import preprocessing

DOC_SET = [
    "weather cold",
    "weather hot dry",
    "weather cold rainny",
    "weather hot",
    "weather cold humid",
]

REFERENCE_STATISTICS_TWO_COMPONENTS = [
    {0: np.array([0.0, 0.0, 0.0]), 1: np.array([0.0, 1.0, 1.0])},
    {0: np.array([0.0, 0.6, 0.0, 0.8, 0.6]), 1: np.array([0.0, 0.4, 0.0, 0.2, 0.4])},
    {
        0: np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
        1: np.array([0.0, 1.0, 1.0, 0.0, 0.0, 1.0]),
    },
    {
        0: np.array([0.0, 0.2, 0.0, 0.6, 0.0, 0.0]),
        1: np.array([0.0, 0.8, 0.0, 0.4, 0.0, 0.0]),
    },
    {
        0: np.array([0.0, 0.0, 0.2, 0.0, 0.0, 0.0, 0.2]),
        1: np.array([0.0, 1.0, 0.8, 0.0, 0.0, 0.0, 0.8]),
    },
]


REFERENCE_STATISTICS_FIVE_COMPONENTS = [
    {
        0: np.array([0.0, 0.4, 0.2]),
        1: np.array([0.0, 0.2, 0.6]),
        2: np.array([0.0, 0.4, 0.0]),
        3: np.array([0.0, 0.0, 0.0]),
        4: np.array([0.0, 0.0, 0.2]),
    },
    {
        0: np.array([0.0, 0.8, 0.0, 0.4, 0.6]),
        1: np.array([0.0, 0.0, 0.0, 0.2, 0.4]),
        2: np.array([0.0, 0.0, 0.0, 0.0, 0.0]),
        3: np.array([0.0, 0.0, 0.0, 0.2, 0.0]),
        4: np.array([0.0, 0.2, 0.0, 0.2, 0.0]),
    },
    {
        0: np.array([0.0, 0.4, 0.2, 0.0, 0.0, 0.0]),
        1: np.array([0.0, 0.2, 0.6, 0.0, 0.0, 0.6]),
        2: np.array([0.0, 0.4, 0.2, 0.0, 0.0, 0.4]),
        3: np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
        4: np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
    },
    {
        0: np.array([0.0, 0.2, 0.0, 0.4, 0.0, 0.0]),
        1: np.array([0.0, 0.2, 0.0, 0.2, 0.0, 0.0]),
        2: np.array([0.0, 0.4, 0.0, 0.2, 0.0, 0.0]),
        3: np.array([0.0, 0.2, 0.0, 0.2, 0.0, 0.0]),
        4: np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
    },
    {
        0: np.array([0.0, 0.0, 0.2, 0.0, 0.0, 0.0, 0.2]),
        1: np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
        2: np.array([0.0, 0.8, 0.8, 0.0, 0.0, 0.0, 0.6]),
        3: np.array([0.0, 0.2, 0.0, 0.0, 0.0, 0.0, 0.2]),
        4: np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
    },
]

REFERENCE_FIVE_COMPONENTS = [
    np.array([1.5, 0.5, 0.5, 0.5, 1.5]),
    np.array([1.5, 1.5, 0.5, 0.5, 1.5]),
    np.array([0.5, 0.5, 3.5, 0.5, 0.5]),
    np.array([0.5, 0.5, 0.5, 2.5, 0.5]),
    np.array([2.5, 0.5, 0.5, 1.5, 0.5]),
]

REFERENCE_COMPONENTS_WITH_PRUNNING = [
    np.array([0.5, 2.5]),
    np.array([3.5, 0.5]),
    np.array([0.5, 3.5]),
    np.array([2.5, 0.5]),
    np.array([2.5, 1.5]),
]

REFERENCE_LEARN_ONE_PREDICT_ONE = [
    np.array([2.5, 0.5]),
    np.array([2.5, 1.5]),
    np.array([1.5, 2.5]),
    np.array([0.5, 2.5]),
    np.array([0.5, 3.5]),
]


def test_extraction_words_ids():
    """
    Assert that input words are split.
    Assert that indexes are updated and extractable.
    """

    lda = preprocessing.LDA(2, number_of_documents=5, seed=42)

    word_indexes_list = []

    for doc in DOC_SET:
        words = doc.split(" ")

        lda._update_indexes(word_list=words)

        word_indexes_list.append([lda.word_to_index[word] for word in words])

    assert word_indexes_list == [[1, 2], [1, 3, 4], [1, 2, 5], [1, 3], [1, 2, 6]]


def test_statistics_two_components():
    """
    Assert that online lda extracts waited statistics on current document.
    """
    n_components = 2

    lda = preprocessing.LDA(n_components, number_of_documents=60, seed=42)

    statistics_list = []

    for doc in DOC_SET:
        word_list = doc.split(" ")

        lda._update_indexes(word_list=word_list)

        word_indexes = [lda.word_to_index[word] for word in word_list]

        statistics, _ = lda._compute_statistics_components(words_indexes_list=word_indexes)

        statistics_list.append(statistics)

        lda._update_weights(statistics=statistics)

    for index, statistics in enumerate(statistics_list):
        for component in range(n_components):
            assert np.array_equal(
                a1=statistics[component],
                a2=REFERENCE_STATISTICS_TWO_COMPONENTS[index][component],
            )


def test_statistics_five_components():
    """
    Assert that online lda extracts waited statistics on current document.
    """

    n_components = 5

    lda = preprocessing.LDA(
        n_components=n_components,
        number_of_documents=60,
        maximum_size_vocabulary=100,
        alpha_beta=100,
        alpha_theta=0.5,
        seed=42,
    )

    statistics_list = []

    for doc in DOC_SET:
        word_list = doc.split(" ")

        lda._update_indexes(word_list=word_list)

        word_indexes = [lda.word_to_index[word] for word in word_list]

        statistics, _ = lda._compute_statistics_components(words_indexes_list=word_indexes)

        statistics_list.append(statistics)

        lda._update_weights(statistics=statistics)

    for index, statistics in enumerate(statistics_list):
        for component in range(n_components):
            assert np.array_equal(
                a1=statistics[component],
                a2=REFERENCE_STATISTICS_FIVE_COMPONENTS[index][component],
            )


def test_five_components():
    """
    Assert that components computed are identical to the original version for n dimensions.
    """

    n_components = 5

    lda = preprocessing.LDA(
        n_components=n_components,
        number_of_documents=60,
        maximum_size_vocabulary=100,
        alpha_beta=100,
        alpha_theta=0.5,
        seed=42,
    )

    components_list = []

    for document in DOC_SET:
        tokens = {token: 1 for token in document.split(" ")}
        components_list.append(lda.learn_transform_one(tokens))

    for index, component in enumerate(components_list):
        assert np.array_equal(a1=list(component.values()), a2=REFERENCE_FIVE_COMPONENTS[index])


def test_prunning_vocabulary():
    """
    Vocabulary prunning is available to improve accuracy and limit memory usage.
    You can perform vocabulary prunning with parameters vocab_prune_interval (int) and
    maximum_size_vocabulary (int).
    """

    lda = preprocessing.LDA(
        n_components=2,
        number_of_documents=60,
        vocab_prune_interval=2,
        maximum_size_vocabulary=3,
        seed=42,
    )

    components_list = []

    for document in DOC_SET:
        tokens = {token: 1 for token in document.split(" ")}
        components_list.append(lda.learn_transform_one(tokens))

    for index, component in enumerate(components_list):
        assert np.array_equal(
            a1=list(component.values()), a2=REFERENCE_COMPONENTS_WITH_PRUNNING[index]
        )


def test_learn_transform():
    """
    Assert that learn_one and transform_one methods returns waited output.
    """

    lda = preprocessing.LDA(
        n_components=2,
        number_of_documents=60,
        vocab_prune_interval=2,
        maximum_size_vocabulary=3,
        seed=42,
    )
    components_list = []

    for document in DOC_SET:
        tokens = {token: 1 for token in document.split(" ")}
        lda.learn_one(x=tokens)

        components_list.append(lda.transform_one(x=tokens))

    for index, component in enumerate(components_list):
        assert np.array_equal(
            a1=list(component.values()), a2=REFERENCE_LEARN_ONE_PREDICT_ONE[index]
        )



from __future__ import annotations

import functools
import importlib
import inspect
import itertools
from urllib import request

import pytest

from river import datasets

from . import base


def _iter_datasets():
    for _, dataset in inspect.getmembers(
        importlib.import_module("river.datasets"), inspect.isclass
    ):
        if issubclass(dataset, datasets.Insects):
            for variant in dataset.variants:
                yield dataset(variant=variant)
            continue
        yield dataset()


@pytest.mark.parametrize(
    "dataset",
    [
        pytest.param(dataset, id=dataset.__class__.__name__)
        for dataset in _iter_datasets()
        if isinstance(dataset, base.RemoteDataset)
    ],
)
@pytest.mark.datasets
def test_remote_url(dataset):
    with request.urlopen(dataset.url) as r:
        assert r.status == 200


@pytest.mark.parametrize(
    "dataset",
    [
        pytest.param(dataset, id=dataset.__class__.__name__)
        for dataset in _iter_datasets()
        if not isinstance(dataset, base.SyntheticDataset)
    ],
)
@pytest.mark.datasets
def test_dimensions(dataset):
    n = 0
    for x, _ in dataset:
        if not dataset.sparse:
            assert len(x) == dataset.n_features
        n += 1
    assert n == dataset.n_samples


@pytest.mark.parametrize(
    "dataset",
    [pytest.param(dataset, id=dataset.__class__.__name__) for dataset in _iter_datasets()],
)
def test_repr(dataset):
    assert repr(dataset)


def _iter_synth_datasets():
    for variant in range(10):
        dataset = functools.partial(datasets.synth.Agrawal, classification_function=variant)
        functools.update_wrapper(dataset, datasets.synth.Agrawal)
        yield dataset

    synth = importlib.import_module("river.datasets.synth")
    for name, dataset in inspect.getmembers(synth, inspect.isclass):
        # TODO: test the following synth datasets also
        if name in ("RandomRBF", "RandomRBFDrift", "RandomTree"):
            continue
        yield dataset


@pytest.mark.parametrize(
    "dataset",
    [pytest.param(dataset(seed=42), id=dataset.__name__) for dataset in _iter_synth_datasets()],
)
def test_synth_idempotent(dataset):
    """Checks that a synthetic dataset produces identical results when seeded."""
    assert list(dataset.take(20)) == list(dataset.take(20))


# HACK: there's no reason Logical should fail this test
@pytest.mark.parametrize(
    "dataset",
    [
        pytest.param(dataset(seed=None), id=dataset.__name__)
        for dataset in _iter_synth_datasets()
        if dataset.__name__ != "Logical"
    ],
)
def test_synth_non_idempotent(dataset):
    """Checks that a synthetic dataset produces different results when not seeded."""
    assert list(dataset.take(20)) != list(dataset.take(20))


@pytest.mark.parametrize(
    "dataset",
    [pytest.param(dataset(seed=42), id=dataset.__name__) for dataset in _iter_synth_datasets()],
)
def test_synth_pausable(dataset):
    stream = iter(dataset)
    s1 = itertools.islice(stream, 3)
    s2 = itertools.islice(stream, 2)
    assert list(dataset.take(5)) == list(itertools.chain(s1, s2))



from __future__ import annotations

import calendar
import math
import random

import pytest
import sympy

from river import compose, datasets, metrics, time_series
from river.time_series.snarimax import Differencer


class Yt(sympy.IndexedBase):
    t = sympy.symbols("t", cls=sympy.Idx)

    def __getitem__(self, idx):
        return super().__getitem__(self.t - idx)


def test_diff_formula():
    """

    >>> import sympy
    >>> from river.time_series.snarimax import Differencer

    >>> Y = Yt('y')
    >>> Y
    y

    >>> p = sympy.symbols('p')
    >>> p
    p

    >>> D = Differencer

    1
    >>> D(0).diff(p, Y)
    p

    (1 - B)
    >>> D(1).diff(p, Y)
    p - y[t]

    (1 - B)^2
    >>> D(2).diff(p, Y)
    p + y[t - 1] - 2*y[t]

    (1 - B^m)
    >>> m = sympy.symbols('m', cls=sympy.Idx)
    >>> D(1, m).diff(p, Y)
    p - y[-m + t + 1]

    (1 - B)(1 - B^m)
    >>> (D(1) * D(1, m)).diff(p, Y)
    p - y[-m + t + 1] + y[-m + t] - y[t]

    (1 - B)(1 - B^12)
    >>> (D(1) * D(1, 12)).diff(p, Y)
    p - y[t - 11] + y[t - 12] - y[t]

    """


def test_diff_example():
    """https://people.duke.edu/~rnau/411sdif.htm

    >>> import pandas as pd
    >>> from river.time_series.snarimax import Differencer

    >>> sales = pd.DataFrame([
    ...     {'date': 'Jan-70', 'autosale': 4.79, 'cpi': 0.297},
    ...     {'date': 'Feb-70', 'autosale': 4.96, 'cpi': 0.298},
    ...     {'date': 'Mar-70', 'autosale': 5.64, 'cpi': 0.300},
    ...     {'date': 'Apr-70', 'autosale': 5.98, 'cpi': 0.302},
    ...     {'date': 'May-70', 'autosale': 6.08, 'cpi': 0.303},
    ...     {'date': 'Jun-70', 'autosale': 6.55, 'cpi': 0.305},
    ...     {'date': 'Jul-70', 'autosale': 6.11, 'cpi': 0.306},
    ...     {'date': 'Aug-70', 'autosale': 5.37, 'cpi': 0.306},
    ...     {'date': 'Sep-70', 'autosale': 5.17, 'cpi': 0.308},
    ...     {'date': 'Oct-70', 'autosale': 5.48, 'cpi': 0.309},
    ...     {'date': 'Nov-70', 'autosale': 4.49, 'cpi': 0.311},
    ...     {'date': 'Dec-70', 'autosale': 4.65, 'cpi': 0.312},
    ...     {'date': 'Jan-71', 'autosale': 5.17, 'cpi': 0.312},
    ...     {'date': 'Feb-71', 'autosale': 5.57, 'cpi': 0.313},
    ...     {'date': 'Mar-71', 'autosale': 6.92, 'cpi': 0.314},
    ...     {'date': 'Apr-71', 'autosale': 7.10, 'cpi': 0.315},
    ...     {'date': 'May-71', 'autosale': 7.02, 'cpi': 0.316},
    ...     {'date': 'Jun-71', 'autosale': 7.58, 'cpi': 0.319},
    ...     {'date': 'Jul-71', 'autosale': 6.93, 'cpi': 0.319},
    ... ])

    >>> sales['autosale/cpi'] = sales.eval('autosale / cpi').round(2)
    >>> Y = sales['autosale/cpi'].to_list()

    >>> diff = Differencer(1)
    >>> sales['(1 - B)'] = [
    ...     diff.diff(p, Y[:i][::-1])
    ...     if i else ''
    ...     for i, p in enumerate(Y)
    ... ]

    >>> sdiff = Differencer(1, 12)
    >>> sales['(1 - B^12)'] = [
    ...     sdiff.diff(p, Y[:i][::-1])
    ...     if i >= 12 else ''
    ...     for i, p in enumerate(Y)
    ... ]

    >>> sales['(1 - B)(1 - B^12)'] = [
    ...     (diff * sdiff).diff(p, Y[:i][::-1])
    ...     if i >= 13 else ''
    ...     for i, p in enumerate(Y)
    ... ]

    >>> sales
          date  autosale    cpi  autosale/cpi (1 - B) (1 - B^12) (1 - B)(1 - B^12)
    0   Jan-70      4.79  0.297         16.13
    1   Feb-70      4.96  0.298         16.64    0.51
    2   Mar-70      5.64  0.300         18.80    2.16
    3   Apr-70      5.98  0.302         19.80     1.0
    4   May-70      6.08  0.303         20.07    0.27
    5   Jun-70      6.55  0.305         21.48    1.41
    6   Jul-70      6.11  0.306         19.97   -1.51
    7   Aug-70      5.37  0.306         17.55   -2.42
    8   Sep-70      5.17  0.308         16.79   -0.76
    9   Oct-70      5.48  0.309         17.73    0.94
    10  Nov-70      4.49  0.311         14.44   -3.29
    11  Dec-70      4.65  0.312         14.90    0.46
    12  Jan-71      5.17  0.312         16.57    1.67       0.44
    13  Feb-71      5.57  0.313         17.80    1.23       1.16              0.72
    14  Mar-71      6.92  0.314         22.04    4.24       3.24              2.08
    15  Apr-71      7.10  0.315         22.54     0.5       2.74              -0.5
    16  May-71      7.02  0.316         22.22   -0.32       2.15             -0.59
    17  Jun-71      7.58  0.319         23.76    1.54       2.28              0.13
    18  Jul-71      6.93  0.319         21.72   -2.04       1.75             -0.53

    """


@pytest.mark.parametrize(
    "differencer",
    [
        Differencer(1),
        Differencer(2),
        Differencer(1, 2),
        Differencer(2, 2),
        Differencer(1, 10),
        Differencer(2, 10),
        Differencer(1) * Differencer(1),
        Differencer(2) * Differencer(1),
        Differencer(1) * Differencer(2),
        Differencer(1) * Differencer(1, 2),
        Differencer(2) * Differencer(1, 2),
        Differencer(1, 2) * Differencer(1, 10),
        Differencer(1, 2) * Differencer(2, 10),
        Differencer(2, 2) * Differencer(1, 10),
        Differencer(2, 2) * Differencer(2, 10),
    ],
)
def test_undiff(differencer):
    Y = [random.random() for _ in range(max(differencer.coeffs))]
    p = random.random()

    diffed = differencer.diff(p, Y)
    undiffed = differencer.undiff(diffed, Y)
    assert math.isclose(undiffed, p)


@pytest.mark.parametrize(
    "snarimax, Y, errors, expected",
    [
        # Non-seasonal parts (p and q)
        (
            time_series.SNARIMAX(p=3, d=0, q=3),
            [1, 2, 3],
            [-4, -5, -6],
            {"e-1": -4, "e-2": -5, "e-3": -6, "y-1": 1, "y-2": 2, "y-3": 3},
        ),
        (
            time_series.SNARIMAX(p=2, d=0, q=3),
            [1, 2, 3],
            [-4, -5, -6],
            {"e-1": -4, "e-2": -5, "e-3": -6, "y-1": 1, "y-2": 2},
        ),
        (
            time_series.SNARIMAX(p=3, d=0, q=2),
            [1, 2, 3],
            [-4, -5, -6],
            {"e-1": -4, "e-2": -5, "y-1": 1, "y-2": 2, "y-3": 3},
        ),
        (
            time_series.SNARIMAX(p=2, d=0, q=2),
            [1, 2, 3],
            [-4, -5, -6],
            {"e-1": -4, "e-2": -5, "y-1": 1, "y-2": 2},
        ),
        # Not enough data, so features too far away are omitted
        (
            time_series.SNARIMAX(p=3, d=0, q=3),
            [1, 2],
            [-4, -5],
            {"e-1": -4, "e-2": -5, "y-1": 1, "y-2": 2},
        ),
        # Seasonal AR
        (
            time_series.SNARIMAX(p=2, d=0, q=2, m=3, sp=2),
            [i for i in range(12)],
            [i for i in range(12)],
            {"e-1": 0, "e-2": 1, "sy-3": 2, "sy-6": 5, "y-1": 0, "y-2": 1},
        ),
        (
            time_series.SNARIMAX(p=2, d=0, q=2, m=2, sp=2),
            [i for i in range(12)],
            [i for i in range(12)],
            {"e-1": 0, "e-2": 1, "sy-2": 1, "sy-4": 3, "y-1": 0, "y-2": 1},
        ),
        (
            time_series.SNARIMAX(p=2, d=0, q=2, m=2, sp=3),
            [i for i in range(12)],
            [i for i in range(12)],
            {"e-1": 0, "e-2": 1, "sy-2": 1, "sy-4": 3, "sy-6": 5, "y-1": 0, "y-2": 1},
        ),
        # Seasonal MA
        (
            time_series.SNARIMAX(p=2, d=0, q=2, m=3, sq=2),
            [i for i in range(12)],
            [i for i in range(12)],
            {"e-1": 0, "e-2": 1, "se-3": 2, "se-6": 5, "y-1": 0, "y-2": 1},
        ),
        (
            time_series.SNARIMAX(p=2, d=0, q=2, m=3, sq=4),
            [i for i in range(12)],
            [i for i in range(12)],
            {"e-1": 0, "e-2": 1, "se-3": 2, "se-6": 5, "se-9": 8, "se-12": 11, "y-1": 0, "y-2": 1},
        ),
        (
            time_series.SNARIMAX(p=1, d=0, q=1, m=2, sq=4),
            [i for i in range(12)],
            [i for i in range(12)],
            {"e-1": 0, "se-2": 1, "se-4": 3, "se-6": 5, "se-8": 7, "y-1": 0},
        ),
    ],
)
def test_add_lag_features(snarimax, Y, errors, expected):
    features = snarimax._add_lag_features(x=None, Y=Y, errors=errors)
    assert features == expected


@pytest.mark.parametrize(
    "snarimax",
    [
        time_series.SNARIMAX(p=1, d=1, q=0, m=12, sp=0, sd=1, sq=0),
        time_series.SNARIMAX(p=0, d=1, q=0, m=12, sp=1, sd=1, sq=0),
        time_series.SNARIMAX(p=1, d=2, q=0, m=12, sp=0, sd=0, sq=0),
        time_series.SNARIMAX(p=1, d=0, q=0, m=12, sp=0, sd=2, sq=0),
    ],
)
def test_no_overflow(snarimax):
    def get_month_distances(x):
        return {
            calendar.month_name[month]: math.exp(-((x["month"].month - month) ** 2))
            for month in range(1, 13)
        }

    def get_ordinal_date(x):
        return {"ordinal_date": x["month"].toordinal()}

    extract_features = compose.TransformerUnion(get_ordinal_date, get_month_distances)

    model = extract_features | snarimax

    time_series.evaluate(
        dataset=datasets.AirlinePassengers(), model=model, metric=metrics.MAE(), horizon=12
    )
